from typing import Any, Dict, Tuple, Union
import torch.fx as fx
from torch.fx.node import Argument, Target, map_aggregate
from torch.fx.proxy import Proxy

from dmx.compressor.sparse import Sparseness


class ConfigurationTransformer(fx.Transformer):
    """
    A transformer that changes the format of the ops according to the cfg file

    Args:
        module (fx.GraphModule): module to be added with additional ops.
        scopeDict (Optional[dict]): Dictionary that maps node name to scope. Defaults to None.
        cfg (Optional[str]): config file for setting the added ops formats. Defaults to None.

    Attributes:
        module (fx.GraphModule): module to be added with additional ops.
        scopeDict (Optional[str]): Dictionary that maps node name to scope.
        cfg (Optional[str]): config file for setting the added ops formats.
    """

    def __init__(self, module: fx.GraphModule, scopeDict: dict, cfg=None):
        super().__init__(module)
        self.config = None
        # if cfg:
        #     self.config = DmxConfig().from_yaml(cfg)
        self.module = module
        self.scopeDict = scopeDict

    def placeholder(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Proxy:
        """
        Execute a ``placeholder`` node. Modifies the format of input_cast according to cfg

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        placeholder_node = self.new_graph.placeholder(target)
        layer = self.scopeDict[placeholder_node.name][0]
        cast_name = target + "_cast"
        if self.config and self.has_module(cast_name):
            layer_key = layer.split("__")[-1]
            if layer_key and layer_key in self.config:
                cast_format = self.config[layer_key]["input_format"]
                self.module.get_submodule(cast_name).set_format(cast_format)
        return super().placeholder(target, args, kwargs)

    def output(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``output`` node. Modifies the format of output_cast according to cfg

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        output_node = self.new_graph.output(target)
        layer = self.scopeDict[output_node.name][0]
        cast_name = target + "_cast"
        if self.config and self.has_module(cast_name):
            layer_key = layer.split("__")[-1]
            if layer_key and layer_key in self.config:
                cast_format = self.config[layer_key]["output_format"]
                self.module.get_submodule(cast_name).set_format(cast_format)
        return super().output(target, args, kwargs)

    def get_attr(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Proxy:
        """
        Execute a ``get_attr`` node. Modifies the format of cast/sparsifier according to cfg

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        get_attr_node = self.new_graph.get_attr(target)
        if get_attr_node.name in self.scopeDict:
            layer = self.scopeDict[get_attr_node.name][0]
            cast_name = target + "_cast"
            sparsify_name = target + "_sparsifier"
            layer_key = layer.split("__")[-1]
            if self.config:
                if layer_key and layer_key in self.config:
                    if "weight" in target:
                        if "weight_format" in self.config[
                            layer_key
                        ] and self.has_module(cast_name):
                            cast_format = self.config[layer_key]["weight_format"]
                            self.module.get_submodule(cast_name).set_format(cast_format)
                        if "weight_sparseness" in self.config[
                            layer_key
                        ] and self.has_module(sparsify_name):
                            sparsify_format = self.config[layer_key][
                                "weight_sparseness"
                            ]
                            self.module.get_submodule(sparsify_name).sparseness = (
                                Sparseness.from_shorthand(sparsify_format)
                            )
                    else:
                        if "bias_format" in self.config[layer_key] and self.has_module(
                            cast_name
                        ):
                            cast_format = self.config[layer_key]["bias_format"]
                            self.module.get_submodule(cast_name).set_format(cast_format)
        return super().get_attr(target, args, kwargs)

    def call_function(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``call_function`` node. Modifies the format of cast according to cfg

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the node and the tracer of the new graph
        """
        call_fnc_node = self.new_graph.call_function(target)
        if self.config and call_fnc_node.name in self.scopeDict:
            layer = self.scopeDict[call_fnc_node.name][0]
            cast_name = call_fnc_node.name + "_cast"
            layer_key = layer.split("__")[-1]
            if (
                layer_key
                and layer_key in self.config
                and self.config[layer_key]["instance"].lower()
                in call_fnc_node.name.lower()
                and self.has_module(cast_name)
            ):
                cast_format = self.config[layer_key]["output_format"]
                self.module.get_submodule(cast_name).set_format(cast_format)
        return super().call_function(target, args, kwargs)

    def transform(self):
        """
        Transform ``self.module`` and return the transformed``GraphModule``.

        Returns:
            A GraphModule with updated ops formats.
        """
        result = super().run()
        if result is not None:

            def strip_proxy(a: Union[Argument, Proxy]) -> Any:
                return a.node if isinstance(a, Proxy) else a

            self.new_graph.output(map_aggregate(result, strip_proxy))
        return self.module

    def has_module(self, target: str) -> bool:
        """
        A function that checks if self.module contains a specific target

        Args:
            target (str): the target to look up

        Returns:
            True if self.module contains a node with the same target
        """
        for name, _ in self.module.named_modules():
            if name == target:
                return True
        return False



import pytest
import torch
from dmx.compressor.modeling import nn as dmxnn


RANDOM_SEED = 0

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
torch.manual_seed(RANDOM_SEED)


@pytest.mark.parametrize("bias", (True, False))
@pytest.mark.parametrize("in_features,out_features", ((8, 8), (64, 64), (1024, 1024)))
@pytest.mark.parametrize("leading_dims", ((1,), (64,), (1, 16), (64, 16)))
def test_linear(leading_dims, in_features, out_features, bias):
    torch_module = torch.nn.Linear(in_features, out_features, bias, device=device)
    dmx_module = dmxnn.Linear(in_features, out_features, bias, device=device)
    dmx_module.weight.data = torch_module.weight.data
    if bias:
        dmx_module.bias.data = torch_module.bias.data
    t_inp = torch.randn(*leading_dims, in_features, device=device).requires_grad_()
    c_inp = t_inp.clone().detach().requires_grad_()
    t_out = torch_module(t_inp)
    c_out = dmx_module(c_inp)
    g_out = torch.randn_like(t_out)
    t_out.backward(g_out)
    c_out.backward(g_out)
    assert torch.allclose(t_out.data, c_out.data, atol=1e-6)
    assert torch.allclose(t_inp.grad, c_inp.grad, atol=1e-6)



import warnings
from typing import Union, Optional, Dict
import torch
import torch.nn as nn
from torch.autograd import Function
from torch.quantization.fake_quantize import FakeQuantize
import transformers
from .format import (
    Format,
    Same,
    FixedPoint,
    BlockFloatingPoint,
)
from .smoothquant import ActivationWeightSmoothQuant
from .observer import ObserverBase, DummyObserver, HistogramObserver
import math
from collections import OrderedDict


class CastToFormat(Function):
    r"""
    A simple STE backward function for numerical cast
    """

    @staticmethod
    def forward(ctx, x, fmt, block_dim):
        ctx.set_materialize_grads(False)
        return fmt.cast(x, block_dim)

    @staticmethod
    def backward(ctx, g):
        return g, None, None

    @staticmethod
    def symbolic(
        g: torch._C.Graph, input: torch._C.Value, fmt: torch._C.Value
    ) -> torch._C.Value:
        if isinstance(fmt, Same):
            return g.op("Identity", input)
        elif isinstance(fmt, BlockFloatingPoint):
            # TODO with dtype for torch > 1.11
            return g.op(
                "com.microsoft::DequantizeBFP",
                *g.op(
                    "com.microsoft::QuantizeBFP",
                    input,
                    bfp_type_i=torch.onnx.symbolic_helper._parse_arg(fmt.bfp_id, "i"),
                    outputs=3,
                ),
                bfp_type_i=torch.onnx.symbolic_helper._parse_arg(fmt.bfp_id, "i"),
                dtype_i=1,
                outputs=1,
            )
        else:
            return None


class CastToDict(torch.nn.ModuleDict):
    def forward(
        self,
        x,
        *args,
        output=False,
        **kwargs,
    ):
        keys = list(self.keys())
        if output:
            if isinstance(x, (tuple, list)):
                return type(x)(self[keys[i]](a) for i, a in enumerate(x))
            return self[keys[0]](x)

        i = 1
        new_args = []
        new_kwargs = {}
        for a in args:
            if isinstance(a, torch.Tensor):
                new_args.append(self[keys[i]](a))
                i += 1
            else:
                new_args.append(a)
        for k, v in kwargs.items():
            if isinstance(v, torch.Tensor):
                new_kwargs[k] = self[k + "_cast"](v)
            else:
                new_kwargs[k] = v
        return self[keys[0]](x), new_args, new_kwargs

    def pack_to_dict(self, param):
        keys = list(self.keys())
        if isinstance(param, (tuple, list)):
            param = {
                keys[i]: p if p is not None else "SAME" for i, p in enumerate(param)
            }
        elif not isinstance(param, dict):
            raise ValueError("format needs to be a dict, tuple or list!")
        if len(param) != len(self):
            warnings.warn(
                f"length of format to set is not equal to length of input_casts, some CastTos might not be set properly!\nlen({param}!={len(self)})"
            )
        return param

    def set_format(self, format: Union[Dict, tuple, list]):
        format = self.pack_to_dict(format)
        for k, f in format.items():
            if isinstance(f, str):
                f = Format.from_shorthand(f)
            if k not in self.keys():
                raise RuntimeError(f"No CastTo with key {k}!")
            self[k].format = f
            if hasattr(self[k], "dtype"):
                self[k].dtype = f
                self[k].activation_post_process.dtype = f

    def disable_fake_quant(self):
        for k in self.keys():
            self[k].disable_fake_quant()

    def enable_fake_quant(self):
        for k in self.keys():
            self[k].enable_fake_quant()

    def enable_observer(self):
        for k in self.keys():
            self[k].enable_observer()

    def disable_observer(self):
        for k in self.keys():
            self[k].disable_observer()


class CastTo(FakeQuantize):
    r"""
    Simulated numerical cast to a target format
    subclassing torch.quantization.fake_quantize.FakeQuantize
    TODO: special state_dict handling to include and exclude flags
    """

    def __init__(
        self,
        format="SAME",
        observer=DummyObserver,
        group_size=None,
        block_dim=-1,
        **fake_quantize_kwargs,
    ):
        self.set_format(format)
        super().__init__(observer=observer, dtype=self.format, **fake_quantize_kwargs)
        if group_size:
            assert torch.ao.quantization.utils.is_per_tensor(
                self.qscheme
            ), "group_size must be used with per tensor quantization scheme"
        self.group_size = group_size if group_size else None
        self.physical_dtype = None
        self.block_dim = block_dim
        self.enable_fake_quant()
        self.disable_observer()

    def set_format(self, format: Union[str, torch.dtype, Format]):
        if isinstance(format, str):
            format = Format.from_shorthand(format)
        self.format = format
        if hasattr(self, "dtype"):
            self.dtype = format
            self.activation_post_process.dtype = format

    def _observer_step(self, x):
        r"""
        Helper method for stepping observer in forward(),
        taken from torch.quantization.fake_quantize.FakeQuantize source
        """
        self.activation_post_process.to(x.device)
        if self.group_size:
            if not hasattr(self, "activation_post_processes"):
                group_num = math.ceil(x.shape[self.ch_axis] * 1.0 / self.group_size)
                self.activation_post_processes = [
                    self.activation_post_process.__class__(
                        dtype=self.format,
                        qscheme=self.qscheme,
                        ch_axis=self.ch_axis,
                    ).to(x.device)
                    for i in range(group_num)
                ]
            xs = torch.split(
                x,
                self.group_size,
                dim=self.ch_axis,
            )
            scale, zero_point = [], []
            mins, maxs = [], []
            for i in range(len(xs)):
                self.activation_post_processes[i](xs[i])
                s, zp = self.activation_post_processes[i].calculate_qparams()
                scale.append(s)
                zero_point.append(zp)
                mins.append(self.activation_post_processes[i].min_val)
                maxs.append(self.activation_post_processes[i].max_val)
            _scale = torch.tensor(scale)
            _zero_point = torch.tensor(zero_point)
            self.activation_post_process.min_val = torch.tensor(mins)
            self.activation_post_process.max_val = torch.tensor(maxs)

        else:
            self.activation_post_process(x.detach().float())
            _scale, _zero_point = self.calculate_qparams()

        _scale, _zero_point = _scale.to(self.scale.device), _zero_point.to(
            self.zero_point.device
        )
        if self.scale.shape != _scale.shape:
            self.scale = torch.zeros_like(_scale)
            self.zero_point = torch.zeros_like(_zero_point)
        self.scale.copy_(_scale)
        self.zero_point.copy_(_zero_point)

    def _get_affine_params(self, x):
        if self.is_per_channel:
            extended_shape = [1] * x.dim()
            channel_dim = x.shape[self.ch_axis]
            extended_shape[self.ch_axis] = channel_dim
            sc = self.scale[:channel_dim].view(extended_shape)
            zp = self.zero_point[:channel_dim].view(extended_shape)
        else:
            sc, zp = self.scale, self.zero_point
        return sc.to(x.device), zp.to(x.device)

    def forward(self, x):
        self.physical_dtype = x.dtype
        if (
            self.observer_enabled[0] == 1
            and x is not None
            and not isinstance(self.format, Same)
        ):
            self._observer_step(x)
        if self.fake_quant_enabled[0] == 1:
            if isinstance(self.format, Format):  # d-Matrix custom format
                if isinstance(self.format, FixedPoint):
                    sc, zp = self._get_affine_params(x)
                    if self.group_size:
                        # duplicate each element in sc and zp for group_size number of times
                        sc = torch.repeat_interleave(sc, self.group_size)[
                            : x.shape[self.ch_axis]
                        ]
                        zp = torch.repeat_interleave(zp, self.group_size)[
                            : x.shape[self.ch_axis]
                        ]
                        sc_shape = [1] * len(x.shape)
                        sc_shape[self.ch_axis] = x.shape[self.ch_axis]
                        sc = sc.view(sc_shape)
                        zp = zp.view(sc_shape)
                    x = x / sc + zp
                x = CastToFormat.apply(x, self.format, self.block_dim)
                if isinstance(self.format, FixedPoint):
                    x = (x - zp) * sc
            else:  # torch.dtype
                x = super().forward(x)
        return x.to(self.physical_dtype)

    def enable_calibration(
        self,
        state: bool = True,
        observer_cls: ObserverBase = HistogramObserver,
        qscheme_to_overload: Optional[torch.qscheme] = None,
        group_size: int = None,
        ch_axis: int = None,
    ) -> None:
        if state: 
            if ch_axis is not None:
                self.ch_axis = (
                    self.activation_post_process.ch_axis
                ) = ch_axis
            if qscheme_to_overload is not None:
                self.qscheme = qscheme_to_overload
                self.is_per_channel = (
                    torch.ao.quantization.utils.is_per_channel(qscheme_to_overload)
                )
            self.group_size = group_size if group_size else None
            if self.group_size:
                assert torch.ao.quantization.utils.is_per_tensor(
                    qscheme_to_overload
                ), "group quantization is to be used with per tensor quantization"
            self.activation_post_process = observer_cls(
                dtype=self.format,
                qscheme=self.qscheme,
                ch_axis=self.ch_axis,
            )
            self.disable_fake_quant()
            self.enable_observer()
        else:
            self.enable_fake_quant()
            self.disable_observer()


    def get_precision(self) -> Optional[int]:
        if isinstance(self.format, (Same, torch.dtype)):
            if self.physical_dtype is not None:
                return torch.finfo(self.physical_dtype).bits
            else:
                raise RuntimeError(
                    "physical_dtype has not been inferred, pass some data through first"
                )
        else:
            return self.format.bit_precision

    def extra_repr(self):
        if self.format.blocked:
            return f"format = dtype = {repr(self.format)}, block_dim = {self.block_dim} \nfake_quant_enabled = {bool(self.fake_quant_enabled)}"
        else:
            return f"format = dtype = {repr(self.format)}, qscheme = {self.qscheme}, ch_axis = {self.ch_axis} \nfake_quant_enabled = {bool(self.fake_quant_enabled)}, observer_enabled = {bool(self.observer_enabled)}, scale = {self.scale.cpu().numpy()}, zero_point = {self.zero_point.cpu().numpy()}, group_size = {self.group_size}"


class Quantize(torch.nn.quantized.Quantize):
    r"""Drop-in replacement of torch.nn.quantized.Quantize
    that supports both torch.dtype and numerical.Format

    Args:
     `scale`: scale of the output Quantized Tensor
     `zero_point`: zero_point of output Quantized Tensor
     `dtype`: data type of output Quantized Tensor, either torch.dtype or Format

    Attributes:
      `scale`, `zero_point`, `dtype`
    """

    def __init__(self, scale, zero_point, dtype: Union[torch.dtype, Format]):
        super().__init__(scale, zero_point, dtype)

    def forward(self, x):
        return torch.ops.dmx.quantize(x)


class DeQuantize(torch.nn.quantized.DeQuantize):
    r"""Drop-in replacement of torch.nn.quantized.DeQuantize
    that supports both torch.dtype and numerical.Format

    Args:
     `scale`: scale of the output Quantized Tensor
     `zero_point`: zero_point of output Quantized Tensor
     `dtype`: data type of output Quantized Tensor, either torch.dtype or Format

    Attributes:
      `scale`, `zero_point`, `dtype`
    """

    def __init__(self, scale=None, zero_point=None, dtype=None):
        super().__init__()

    def forward(self, x):
        return torch.ops.dmx.dequantize(x)


class NumericalCastMixin:
    r"""
    Mixin for modules with boundary casting
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.align_boundary_dtype = True
        self.infer_ch_axis()
        self.init_casts()
        self.init_smoothquant()

    def infer_ch_axis(self):
        if isinstance(
            self,
            (nn.Linear,),
        ):
            self.ch_axis = -1
            self.win_ch_axis = -1
            self.wout_ch_axis = 0
        elif isinstance(
            self,
            (transformers.pytorch_utils.Conv1D,),
        ):
            self.ch_axis = -1
            self.win_ch_axis = 0
            self.wout_ch_axis = -1
        elif isinstance(
            self,
            (nn.modules.conv._ConvNd,),
        ):
            self.ch_axis = 1
            self.win_ch_axis = 1
            self.wout_ch_axis = 0
        else:
            self.ch_axis = self.win_ch_axis = self.wout_ch_axis = None

    def init_casts(self) -> None:
        # dynamic i/o casts
        self.input_casts = CastToDict(
            OrderedDict({"input_cast": CastTo(ch_axis=self.ch_axis)})
        )
        self.output_casts = CastToDict(OrderedDict({"output_cast": CastTo()}))
        # dynamic intermediate casts
        if isinstance(
            self,
            (
                nn.Linear,
                nn.Bilinear,
                nn.EmbeddingBag,
                nn.modules.conv._ConvNd,
            ),
        ):
            self.accum_cast = CastTo()
        else:
            self.accum_cast = None
        # static parameter casts
        pnames = [n for n, _ in self.named_parameters()]
        self.weight_storage_cast = (
            CastTo(ch_axis=self.wout_ch_axis) if "weight" in pnames else None
        )
        self.weight_cast = (
            CastTo(ch_axis=self.wout_ch_axis) if "weight" in pnames else None
        )
        self.bias_cast = CastTo() if "bias" in pnames else None
        self.residual_cast = None
        self.multiplier_cast = None

    def init_smoothquant(
        self,
        migration_strength: float = 0.5,
        scale_format: Union[str, Format] = "SAME",
        dynamic: bool = False,
    ) -> None:
        self.smoothquant = (
            ActivationWeightSmoothQuant(
                self.ch_axis,
                self.win_ch_axis,
                migration_strength,
                scale_format,
                dynamic,
            )
            if self.ch_axis is not None and self.win_ch_axis is not None
            else None
        )

    @staticmethod
    def _check_format_dim_consistency(format: Format, ch_axis: int) -> bool:
        return not format.blocked or format.block_dim == ch_axis

    def check_weight_format_dim_consistency(self) -> bool:
        _good = self.weight_cast is None or self._check_format_dim_consistency(
            self.weight_cast.format, self.win_ch_axis
        )
        if not _good:
            warnings.warn(
                f"layer's weight input channel axis {self.win_ch_axis} might be inconsistent with format {self.weight_format}",
                RuntimeWarning,
            )
        return _good

    def check_input_format_dim_consistency(self) -> bool:
        _good = (
            self.input_casts is None
            or self.input_casts.input_cast is None
            or self._check_format_dim_consistency(
                self.input_casts.input_cast.format, self.ch_axis
            )
        )
        if not _good:
            warnings.warn(
                f"Input format dim consistency check failed! Layer's input channel axis {self.ch_axis} might be inconsistent with format {self.input_formats}",
                RuntimeWarning,
            )
        return _good

    def check_residual_format_dim_consistency(self) -> bool:
        _good = (
            self.input_casts is None
            or self.input_casts.residual_cast is None
            or self._check_format_dim_consistency(
                self.input_casts.residual_cast.format, self.ch_axis
            )
        )
        if not _good:
            warnings.warn(
                f"Residual format dim consistency check failed! Layer's input channel axis {self.ch_axis} might be inconsistent with format {self.input_formats}",
                RuntimeWarning,
            )
        return _good

    def check_format_dim_consistency(self) -> bool:
        return (
            self.check_input_format_dim_consistency()
            and self.check_residual_format_dim_consistency()
            and self.check_weight_format_dim_consistency()
        )

    @property
    def input_formats(self):
        return {k: cast.format for k, cast in self.input_casts.items()}

    @property
    def input_precision(self):
        return self.input_casts.input_cast.get_precision()

    @property
    def weight_precision(self):
        return self.weight_cast.get_precision()

    @property
    def weight_storage_precision(self):
        return self.weight_storage_cast.get_precision()

    @property
    def weight_scale(self):
        return self.weight_cast.scale.to(self.weight.device)

    @property
    def weight_zero_point(self):
        return self.weight_cast.zero_point.to(self.weight.device)

    @property
    def weight_storage_scale(self):
        return self.weight_storage_cast.scale.to(self.weight.device)

    @property
    def weight_storage_zero_point(self):
        return self.weight_storage_cast.zero_point.to(self.weight.device)

    @property
    def output_formats(self):
        return {k: cast.format for k, cast in self.output_casts.items()}

    @property
    def accum_format(self):
        return (self.accum_cast.format) if self.accum_cast is not None else None

    @property
    def weight_format(self):
        return (self.weight_cast.format) if self.weight_cast is not None else None

    @property
    def weight_storage_format(self):
        return (
            (self.weight_storage_cast.format)
            if self.weight_storage_cast is not None
            else None
        )

    @property
    def residual_format(self):
        return (self.residual_cast.format) if self.residual_cast is not None else None

    @property
    def multiplier_format(self):
        return (
            (self.multiplier_cast.format) if self.multiplier_cast is not None else None
        )

    @property
    def bias_format(self):
        return (self.bias_cast.format) if self.bias_cast is not None else None



import math
from typing import Union, List, Optional
from collections import OrderedDict
import torch
from torch import Tensor, Size
import torch.nn.functional as F
from .core import DmxGraph
from torch.fx import Graph
import transformers

from dmx.compressor.numerical import Same, CastTo, CastToDict
from . import DmxModule


class ResAdd(DmxModule, torch.nn.Module):
    """
    A module for handling residual connections.
    """

    def __init__(self) -> None:
        super().__init__()
        self.input_casts = CastToDict(
            OrderedDict({"input_cast": CastTo(), "residual_cast": CastTo()})
        )

    def _forward(self, _input: Tensor, _residual: Tensor) -> Tensor:
        """
        A forward pass of addition operation with quantization applied

        Args:
            _input (Tensor): already quantized input tensor
            residual (Tensor): residual tensor

        Returns:
            Sum of _input tensor and quantized residual tensor.
        """
        _output = _input + _residual
        return _output

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            input_dq, residual_dq = g.create_placeholders(
                ["_input", "residual"],
                ["input_casts.input_cast", "input_casts.residual_cast"],
                [
                    repr(self.input_casts.input_cast.format),
                    repr(self.input_casts.residual_cast.format),
                ],
            )

            _output_dq = g.create_node(
                "call_function",
                torch.add,
                (input_dq, residual_dq),
                name="output",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class Mul(DmxModule):
    def __init__(self) -> None:
        super().__init__()
        self.input_casts = CastToDict(
            OrderedDict(
                {
                    "input_cast": CastTo(block_dim=-1),
                    "multiplier_cast": CastTo(block_dim=-2),
                }
            )
        )

    def _forward(self, _input: Tensor, multiplier: Tensor) -> Tensor:
        return _input * multiplier

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq, multiplier_dq = g.create_placeholders(
                ["_input", "multiplier"],
                ["input_casts.input_cast", "input_casts.multiplier_cast"],
                [
                    repr(self.input_casts.input_cast.format),
                    repr(self.input_casts.multiplier_cast.format),
                ],
            )
            _output_dq = g.create_node(
                "call_function",
                torch.mul,
                (_input_dq, multiplier_dq),
                name="output",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class ScaledDotProductAttention(DmxModule):
    is_compound = True

    def __init__(self, dropout_p=0.0) -> None:
        super().__init__()
        self.input_casts = CastToDict(
            OrderedDict(
                {
                    "query_states_cast": CastTo(block_dim=-1),
                    "key_states_cast": CastTo(block_dim=-1),
                    "value_states_cast": CastTo(block_dim=-1),
                    "attn_mask_cast": CastTo(block_dim=-1),
                }
            )
        )
        self.resadd = ResAdd()
        self.matmul = ActActMatMul()
        self.softmax = Softmax(dim=-1)
        self.dropout = Dropout(p=dropout_p)
        self.mul = Mul()

    def forward(
        self,
        query,
        key,
        value,
        attn_mask=None,
        is_causal=False,
        scale=None,
        enable_gqa=False,
    ):
        L, S = query.size(-2), key.size(-2)
        scale_factor = 1 / math.sqrt(query.size(-1)) if scale is None else scale
        attn_bias = torch.zeros(L, S, dtype=query.dtype).to(query.device)

        if is_causal:
            assert attn_mask is None
            temp_mask = torch.ones(L, S, dtype=torch.bool).tril(diagonal=0)
            attn_bias.masked_fill_(temp_mask.logical_not(), float("-inf"))
            attn_bias.to(query.dtype)

        if attn_mask is not None:
            if attn_mask.dtype == torch.bool:
                attn_bias.masked_fill_(attn_mask.logical_not(), float("-inf"))
            else:
                attn_bias = self.resadd(attn_bias, attn_mask)

        if enable_gqa:
            key = key.repeat_interleave(query.size(-3) // key.size(-3), -3)
            value = value.repeat_interleave(query.size(-3) // value.size(-3), -3)

        attn_weight = self.matmul(query, self.mul(key.transpose(-2, -1), scale_factor))
        attn_weight = self.resadd(attn_weight, attn_bias)
        attn_weight = self.softmax(attn_weight)
        attn_weight = self.dropout(attn_weight)
        return self.matmul(attn_weight, value)

    def module_graph(self, *args, **kwargs) -> Graph:
        from dmx.compressor.fx.transform import prepare_tracing_inputs
        from dmx.compressor.fx.tracer import hf_symbolic_trace

        input_names, concrete_args, dummy_inputs = prepare_tracing_inputs(
            self, args, kwargs
        )
        gm, tracer = hf_symbolic_trace(
            self,
            input_names,
            concrete_args=concrete_args,
            dummy_inputs=dummy_inputs,
        )
        return gm

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        raise NotImplementedError(
            "ScaledDotProductAttention is a compound module and does not support to_compiler_graph"
        )


class ActActMatMul(DmxModule, torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.input_casts = CastToDict(
            OrderedDict(
                {
                    "input_cast": CastTo(block_dim=-1),
                    "multiplier_cast": CastTo(block_dim=-2),
                }
            )
        )

    def _forward(self, _input: Tensor, _multiplier: Tensor) -> Tensor:
        _output = torch.matmul(_input, _multiplier)
        return _output

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq, multiplier_dq = g.create_placeholders(
                ["_input", "multiplier"],
                ["input_casts.input_cast", "input_casts.multiplier_cast"],
                [
                    repr(self.input_casts.input_cast.format),
                    repr(self.input_casts.multiplier_cast.format),
                ],
            )
            _output_dq = g.create_node(
                "call_function",
                torch.matmul,
                (_input_dq, multiplier_dq),
                name="output",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class Exp(DmxModule, torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def _forward(self, _input: Tensor) -> Tensor:
        _output = torch.exp(_input)
        return _output

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            _output_dq = g.create_node(
                "call_function",
                torch.exp,
                (_input_dq,),
                name="output",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class BAddBMM(DmxModule):
    def __init__(self) -> None:
        super().__init__()
        self.input_casts = CastToDict(
            OrderedDict(
                {
                    "input_cast": CastTo(block_dim=-1),
                    "batch1_cast": CastTo(block_dim=-1),
                    "batch2_cast": CastTo(block_dim=-2),
                }
            )
        )

    def _forward(self, input, batch1, batch2, **kwargs):
        return torch.baddbmm(input, batch1, batch2, **kwargs)

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            dqs = g.create_placeholders(
                ["_input", "batch1", "batch2"],
                [
                    "input_casts.input_cast",
                    "input_casts.batch1_cast",
                    "input_casts.batch2_cast",
                ],
                [
                    repr(self.input_casts.input_cast.format),
                    repr(self.input_casts.batch1_cast.format),
                    repr(self.input_casts.batch2_cast.format),
                ],
            )
            _output_dq = g.create_node(
                "call_function",
                torch.baddbmm,
                tuple(dqs),
                name="output",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class Linear(DmxModule, torch.nn.Linear):
    r"""
    An extension of PyTorch's Linear layer to support DmxModule configurations.
    This module performs a linear transformation on the input data.

    Args:
        in_features (int): Size of each input sample.
        out_features (int): Size of each output sample.
        bias (bool, optional): If set to False, the layer will not learn an additive bias. Defaults to True.
        **kwargs: Additional keyword arguments inherited from DmxModule.

    Attributes:
        _weight (Tensor): The learnable weights of the module of shape (out_features, in_features).
        _bias (Tensor): The learnable bias of the module of shape (out_features).

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the linear layer.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(in_features, out_features, bias=bias, **kwargs)
        self.input_casts.input_cast.block_dim = -1
        self.weight_cast.block_dim = -1
        if self.bias_cast is not None:
            self.bias_cast.block_dim = -1

    def _forward(self, _input: Tensor) -> Tensor:
        if isinstance(self.accum_format, Same):
            _weight = self._weight.to(_input.dtype)
            _bias = None if self._bias is None else self._bias.to(_input.dtype)
            _output = torch.nn.functional.linear(_input, _weight, _bias)
        else:
            _weight = self._weight
            _product = self.accum_cast(
                torch.matmul(_input.to(_weight.dtype), _weight.t())
            )
            if self.bias is not None:
                _output = torch.add(_product, self._bias)
            else:
                _output = _product
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        if isinstance(raw, transformers.pytorch_utils.Conv1D):
            initial_dmx = cls(
                raw.weight.shape[0], raw.weight.shape[1], bias=raw.bias is not None
            )
            initial_dmx.weight.data = raw.weight.data.t()
            initial_dmx.bias = raw.bias
        else:
            initial_dmx = cls(
                raw.in_features, raw.out_features, bias=raw.bias is not None
            )
            initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> DmxGraph:
        """
        Returns a compiler friendly graph

        >>> Reference:
            opcode         name                               target                             args                                                                               kwargs
            -------------  ---------------------------------  ---------------------------------  ---------------------------------------------------------------------------------  --------
            placeholder    _input                             _input                             ()                                                                                 {}
            get_attr       input_casts_input_cast_scale       input_casts.input_cast.scale       ()                                                                                 {}
            get_attr       input_casts_input_cast_zero_point  input_casts.input_cast.zero_point  ()                                                                                 {}
            call_function  quantize                           dmx.quantize                       (_input, input_casts_input_cast_scale, input_casts_input_cast_zero_point, 'SAME')  {}
            call_function  dequantize                         dmx.dequantize                     (quantize, input_casts_input_cast_scale, input_casts_input_cast_zero_point)        {}
            get_attr       _weight                            _weight                            ()                                                                                 {}
            get_attr       weight_scale                       weight_scale                       ()                                                                                 {}
            get_attr       weight_zero_point                  weight_zero_point                  ()                                                                                 {}
            call_function  quantize_1                         dmx.quantize                       (_weight, weight_scale, weight_zero_point, 'SAME')                                 {}
            call_function  dequantize_1                       dmx.dequantize                     (quantize_1, weight_scale, weight_zero_point)                                      {}
            get_attr       _bias                              _bias                              ()                                                                                 {}
            get_attr       bias_cast_scale                    bias_cast.scale                    ()                                                                                 {}
            get_attr       bias_cast_zero_point               bias_cast.zero_point               ()                                                                                 {}
            call_function  quantize_2                         dmx.quantize                       (_bias, bias_cast_scale, bias_cast_zero_point, 'SAME')                             {}
            call_function  dequantize_2                       dmx.dequantize                     (quantize_2, bias_cast_scale, bias_cast_zero_point)                                {}
            call_function  _output                            <built-in function linear>         (dequantize, dequantize_1, dequantize_2)                                           {}
            get_attr       output_casts_output_cast_scale     output_casts.output_cast.scale     ()                                                                                 {}
            get_attr       output_casts_output_cast_zero_pointoutput_casts.utput_cast.zero_point ()                                                                                 {}
            call_function  quantize_3                         dmx.quantize                       (_output, output_casts_output_cast_scale, output_casts_output_cast_zero_pointoutput_casts, 'SAME')  {}
            call_function  dequantize_3                       dmx.dequantize                     (quantize_3, output_casts_output_cast_scale, output_casts_output_cast_zero_pointoutput_casts)       {}
            output         output                             output                             (dequantize_3,)                                                                    {}

        """
        g = DmxGraph()
        with g.inserting_after():
            # PLACEHOLDERS
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )

            # _weight
            _weight_storage_dq = g.get_attr(
                "weight", "weight_storage_cast", repr(self.weight_storage_cast.format)
            )
            _weight_dq = g.qdq_node(
                _weight_storage_dq, "weight_cast", repr(self.weight_cast.format)
            )

            # _bias
            _bias_dq = g.get_attr(
                "bias",
                "bias_cast",
                repr(getattr(self.bias_cast, "format", None)),
                optional_arg=self.bias,
            )

            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.linear,
                (_input_dq, _weight_dq, _bias_dq),
                name="_output",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class Embedding(DmxModule, torch.nn.Embedding):
    r"""
    An extension of PyTorch's Embedding layer to support DmxModule configurations.

    Args:
        num_embeddings (int): size of the dictionary of embeddings
        embedding_dim (int): the size of each embedding vector
        **kwargs: Additional keyword arguments inherited from torch.nn.Embedding.

    Attributes:
        _weight (Tensor):the learnable weights of the module of shape (num_embeddings, embedding_dim).

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the embedding layer.
    """

    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        **kwargs,
    ) -> None:
        super().__init__(num_embeddings, embedding_dim, **kwargs)
        self.align_boundary_dtype = False  # special treatment for sparse layers

    def _forward(self, _input: Tensor) -> Tensor:
        _output = F.embedding(
            _input,
            self._weight,
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse,
        )
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        initial_dmx = cls(
            num_embeddings=raw.num_embeddings,
            embedding_dim=raw.embedding_dim,
            padding_idx=raw.padding_idx,
            max_norm=raw.max_norm,
            norm_type=raw.norm_type,
            scale_grad_by_freq=raw.scale_grad_by_freq,
            sparse=raw.sparse,
        )
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            # PLACEHOLDERS
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )

            # _weight
            _weight_dq = g.get_attr(
                "weight", "weight_cast", repr(self.weight_cast.format)
            )
            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.embedding,
                (
                    _input_dq,
                    _weight_dq,
                    self.padding_idx,
                    self.max_norm,
                    self.norm_type,
                    self.scale_grad_by_freq,
                    self.sparse,
                ),
                name="_output",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class Conv1d(DmxModule, torch.nn.Conv1d):
    r"""
    An extension of PyTorch's Conv1d layer to support DmxModule configurations.
    This module performs a 1D convolution over an input signal composed of several input planes.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the convolution.
        kernel_size (int or tuple): Size of the convolving kernel.
        stride (int or tuple, optional): Stride of the convolution. Defaults to 1.
        padding (int or tuple, optional): Zero-padding added to both sides of the input. Defaults to 0.
        dilation (int or tuple, optional): Spacing between kernel elements. Defaults to 1.
        groups (int, optional): Number of blocked connections from input channels to output channels. Defaults to 1.
        bias (bool, optional): If set to False, the layer will not learn an additive bias. Defaults to True.
        padding_mode (str, optional): Accepted values 'zeros' and 'circular' etc. Defaults to 'zeros'.
        **kwargs: Additional keyword arguments inherited from DmxModule.

    Attributes:
        _weight (Tensor): The learnable weights of the module of shape (out_channels, in_channels, kernel_size).
        _bias (Tensor, optional): The learnable bias of the module of shape (out_channels).

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the 1D convolution.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        **kwargs,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            **kwargs,
        )
        self.input_casts.input_cast.block_dim = 1
        self.weight_cast.block_dim = 1
        if self.bias_cast is not None:
            self.bias_cast.block_dim = -1

    def _forward(self, _input: Tensor) -> Tensor:
        _weight = self._weight
        _convolution = self.accum_cast(
            self._conv_forward(_input.to(_weight.dtype), _weight, None)
        )
        if self.bias is not None:
            _output = torch.add(_convolution, self._bias.unsqueeze(-1))
        else:
            _output = _convolution
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        """
        Creates a new Conv1d object (DmxModule) from a given PyTorch Conv1d layer.

        Args:
            raw (torch.nn.Module): A PyTorch Conv1d layer to be converted.

        Returns:
            DmxModule: A Conv1d object that has the same configuration as the input PyTorch Conv1d layer.
        """
        initial_dmx = cls(
            raw.in_channels,
            raw.out_channels,
            raw.kernel_size,
            stride=raw.stride,
            padding=raw.padding,
            dilation=raw.dilation,
            groups=raw.groups,
            bias=raw.bias is not None,
            padding_mode=raw.padding_mode,
        )
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx


class Conv2d(DmxModule, torch.nn.Conv2d):
    r"""
    An extension of PyTorch's Conv2d layer to support DmxModule configurations.
    This module performs a 2D convolution over an input signal composed of several input planes.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the convolution.
        kernel_size (int or tuple): Size of the convolving kernel.
        stride (int or tuple, optional): Stride of the convolution. Defaults to 1.
        padding (int or tuple, optional): Zero-padding added to both sides of the input. Defaults to 0.
        dilation (int or tuple, optional): Spacing between kernel elements. Defaults to 1.
        groups (int, optional): Number of blocked connections from input channels to output channels. Defaults to 1.
        bias (bool, optional): If set to False, the layer will not learn an additive bias. Defaults to True.
        padding_mode (str, optional): Accepted values 'zeros' and 'circular' etc. Defaults to 'zeros'.
        **kwargs: Additional keyword arguments inherited from DmxModule.

    Attributes:
        _weight (Tensor): The learnable weights of the module of shape (out_channels, in_channels, kernel_height, kernel_width).
        _bias (Tensor, optional): The learnable bias of the module of shape (out_channels).

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the 2D convolution.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        **kwargs,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            **kwargs,
        )
        self.input_casts.input_cast.block_dim = 1
        self.weight_cast.block_dim = 1
        if self.bias_cast is not None:
            self.bias_cast.block_dim = -1

    def _forward(self, _input: Tensor) -> Tensor:
        _weight = self._weight
        _convolution = self.accum_cast(
            self._conv_forward(_input.to(_weight.dtype), _weight, None)
        )
        if self.bias is not None:
            _output = torch.add(_convolution, self._bias.unsqueeze(-1).unsqueeze(-1))
        else:
            _output = _convolution
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        """
        Creates a new Conv2d object (DmxModule) from a given PyTorch Conv2d layer.

        Args:
            raw (torch.nn.Module): A PyTorch Conv2d layer to be converted.

        Returns:
            DmxModule: A Conv2d object that has the same configuration as the input PyTorch Conv2d layer.
        """
        initial_dmx = cls(
            raw.in_channels,
            raw.out_channels,
            raw.kernel_size,
            stride=raw.stride,
            padding=raw.padding,
            dilation=raw.dilation,
            groups=raw.groups,
            bias=raw.bias is not None,
            padding_mode=raw.padding_mode,
        )
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx


class ConvTranspose2d(DmxModule, torch.nn.ConvTranspose2d):
    r"""
    An extension of PyTorch's ConvTranspose2d layer to support DmxModule configurations.
    This module performs a 2D transposed convolution over an input signal composed of several input planes.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the transposed convolution.
        kernel_size (int or tuple): Size of the convolving kernel.
        stride (int or tuple, optional): Stride of the transposed convolution. Defaults to 1.
        padding (int or tuple, optional): Zero-padding added to both sides of the input. Defaults to 0.
        output_padding (int or tuple, optional): Additional size added to one side of the output shape. Defaults to 0.
        groups (int, optional): Number of blocked connections from input channels to output channels. Defaults to 1.
        bias (bool, optional): If set to False, the layer will not learn an additive bias. Defaults to True.
        padding_mode (str, optional): Accepted values 'zeros'. Defaults to 'zeros'.
        **kwargs: Additional keyword arguments inherited from DmxModule.

    Attributes:
        _weight (Tensor): The learnable weights of the module of shape (in_channels, out_channels, kernel_height, kernel_width).
        _bias (Tensor, optional): The learnable bias of the module of shape (out_channels).

    Methods:
        _forward (_input: Tensor, output_size: Optional[List[int]] = None) -> Tensor: Computes the forward pass of the 2D transposed convolution.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        output_padding=0,
        groups=1,
        bias=True,
        padding_mode="zeros",
        **kwargs,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            **kwargs,
        )
        self.input_casts.input_cast.block_dim = 1
        self.weight_cast.block_dim = 1
        if self.bias_cast is not None:
            self.bias_cast.block_dim = -1

    def _forward(
        self, _input: Tensor, output_size: Optional[List[int]] = None
    ) -> Tensor:
        if self.padding_mode != "zeros":
            raise ValueError(
                "Only `zeros` padding mode is supported for ConvTranspose2d"
            )

        assert isinstance(self.padding, tuple)
        # One cannot replace List by Tuple or Sequence in "_output_padding" because
        # TorchScript does not support `Sequence[T]` or `Tuple[T, ...]`.
        num_spatial_dims = 2
        output_padding = self._output_padding(
            _input,
            output_size,
            self.stride,
            self.padding,
            self.kernel_size,  # type: ignore[arg-type]
            num_spatial_dims,
            self.dilation,
        )  # type: ignore[arg-type]

        _weight = self._weight
        _convolution = self.accum_cast(
            F.conv_transpose2d(
                _input.to(_weight.dtype),
                _weight,
                None,
                self.stride,
                self.padding,
                output_padding,
                self.groups,
                self.dilation,
            )
        )
        if self.bias is not None:
            _output = torch.add(_convolution, self._bias.unsqueeze(-1).unsqueeze(-1))
        else:
            _output = _convolution
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        initial_dmx = cls(
            raw.in_channels,
            raw.out_channels,
            raw.kernel_size,
            stride=raw.stride,
            padding=raw.padding,
            dilation=raw.dilation,
            groups=raw.groups,
            bias=raw.bias is not None,
            padding_mode=raw.padding_mode,
        )
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx


class AdaptiveAvgPool2d(DmxModule, torch.nn.AdaptiveAvgPool2d):
    r"""
    An extension of PyTorch's AdaptiveAvgPool2d layer to support DmxModule configurations.
    This module applies a 2D adaptive average pooling over an input signal composed of several input planes.

    Args:
        output_size (int or tuple): The size of the output tensor after pooling.

    Attributes:
        None specific to this subclass. Inherits attributes from parent classes.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the 2D adaptive average pooling.
    """

    def __init__(self, output_size) -> None:
        super().__init__(output_size)

    def _forward(self, _input: Tensor) -> Tensor:
        _output = self.approx_forward((_input,))
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        """
        Creates a new AdaptiveAvgPool2d object (DmxModule) from a given PyTorch AdaptiveAvgPool2d layer.

        Args:
            raw (torch.nn.Module): A PyTorch AdaptiveAvgPool2d layer to be converted.

        Returns:
            DmxModule: An AdaptiveAvgPool2d object that has the same configuration as the input PyTorch AdaptiveAvgPool2d layer.
        """
        initial_dmx = cls(raw.output_size)
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx


class AvgPool2d(DmxModule, torch.nn.AvgPool2d):
    r"""
    An extension of PyTorch's AvgPool2d layer to support DmxModule configurations.
    This module applies a 2D average pooling over an input signal composed of several input planes.

    Args:
        output_size (int or tuple): The size of the output tensor after pooling.

    Attributes:
        None specific to this subclass. Inherits attributes from parent classes.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the 2D average pooling.
    """

    def __init__(self, output_size) -> None:
        super().__init__(output_size)

    def _forward(self, _input: Tensor) -> Tensor:
        _output = self.approx_forward((_input,))
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        """
        Creates a new AvgPool2d object (DmxModule) from a given PyTorch AvgPool2d layer.

        Args:
            raw (torch.nn.Module): A PyTorch AvgPool2d layer to be converted.

        Returns:
            DmxModule: An AvgPool2d object that has the same configuration as the input PyTorch AvgPool2d layer.
        """
        initial_dmx = cls(raw.output_size)
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx


class MaxPool2d(DmxModule, torch.nn.MaxPool2d):
    r"""
    An extension of PyTorch's MaxPool2d layer to support DmxModule configurations.
    This module applies a 2D max pooling over an input signal composed of several input planes.

    Args:
        kernel_size (int or tuple): Size of the window to take a max over.
        stride (int or tuple, optional): Stride of the window. Defaults to None.
        padding (int or tuple, optional): Zero-padding added to both sides of the input. Defaults to 0.
        dilation (int or tuple, optional): Spacing between kernel elements. Defaults to 1.
        return_indices (bool, optional): If True, will return the max indices in a second tensor. Defaults to False.
        ceil_mode (bool, optional): If True, will use ceil instead of floor to compute the output shape. Defaults to False.

    Attributes:
        None specific to this subclass. Inherits attributes from parent classes.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the 2D max pooling.
    """

    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        dilation=1,
        return_indices=False,
        ceil_mode=False,
    ) -> None:
        super().__init__(
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            return_indices=return_indices,
            ceil_mode=ceil_mode,
        )

    def _forward(self, _input: Tensor) -> Tensor:
        _output = self.approx_forward((_input,))
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        """
        Creates a new MaxPool2d object (DmxModule) from a given PyTorch MaxPool2d layer.

        Args:
            raw (torch.nn.Module): A PyTorch MaxPool2d layer to be converted.

        Returns:
            DmxModule: A MaxPool2d object that has the same configuration as the input PyTorch MaxPool2d layer.
        """
        initial_dmx = cls(
            raw.kernel_size,
            stride=raw.stride,
            padding=raw.padding,
            dilation=raw.dilation,
            return_indices=raw.return_indices,
            ceil_mode=raw.ceil_mode,
        )
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx


class Softmax(DmxModule, torch.nn.Softmax):
    r"""
    An extension of PyTorch's Softmax layer to support DmxModule configurations.
    This module applies the Softmax function to an n-dimensional input tensor, normalizing the elements
    along a specified dimension such that they sum up to 1.

    Args:
        dim (int, optional): Dimension along which Softmax will be computed. Defaults to -1.

    Attributes:
        None specific to this subclass. Inherits attributes from parent classes.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the Softmax function.
    """

    def __init__(self, dim: int = -1) -> None:
        super().__init__(dim=dim)

    def _forward(self, _input: Tensor, *args, **kwargs) -> Tensor:
        _output = self.approx_forward((_input,), dim=self.dim)
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        """
        Creates a Softmax DmxModule instance from a raw torch.nn.Module instance.

        Args:
            raw (torch.nn.Module): The raw torch.nn.Module instance.

        Returns:
            DmxModule: An initialized Softmax DmxModule instance with parameters copied from the raw instance.
        """
        initial_dmx = cls(dim=raw.dim)
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            dim = g.get_attr("dim")
            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.softmax,
                (_input_dq, dim),
                name="softmax",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class LayerNorm(DmxModule, torch.nn.LayerNorm):
    r"""
    An extension of PyTorch's LayerNorm layer to support DmxModule configurations.
    This module applies layer normalization over dimensions specified by the `normalized_shape` attribute.
    The mean and standard deviation are computed over the last `D` dimensions, where `D`is the dimensionality indicated by `normalized_shape`.
    Gamma and Beta are learnable parameters if `elementwise_affine` is set to True.

    Args:
        normalized_shape (Union[int, List[int], Size]): Specifies dimensions for layer normalization.
        eps (float, optional): A value added for numerical stability. Defaults to 1e-5.
        elementwise_affine (bool, optional): Indicates if learnable affine parameters Gamma and Beta should be used. Defaults to True.

    Attributes:
        None specific to this subclass. Inherits attributes from parent classes.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the layer normalization.
    """

    def __init__(
        self,
        normalized_shape: Union[int, List[int], Size],
        eps: float = 1e-5,
        elementwise_affine: bool = True,
    ) -> None:
        super().__init__(
            normalized_shape, eps=eps, elementwise_affine=elementwise_affine
        )
        self.functional_forward = F.layer_norm

    def _forward(self, _input: Tensor, *args, **kwargs) -> Tensor:
        _output = self.approx_forward(
            (_input,), self.normalized_shape, self._weight, self._bias, self.eps
        )
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        """
        Creates a new LayerNorm object (DmxModule) from a given PyTorch LayerNorm layer.

        Args:
            raw (torch.nn.Module): A PyTorch LayerNorm layer to be converted.

        Returns:
            DmxModule: A LayerNorm object that has the same configuration as the input PyTorch LayerNorm layer.
        """
        initial_dmx = cls(
            raw.normalized_shape, eps=raw.eps, elementwise_affine=raw.elementwise_affine
        )
        initial_dmx.update_params_with_raw(raw)
        initial_dmx.type(raw.weight.dtype)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )

            # Tensor Attributes

            _weight_dq = g.get_attr(
                "weight", "weight_cast", repr(self.weight_cast.format)
            )

            _bias_dq = g.get_attr(
                "bias",
                "bias_cast",
                repr(getattr(self.bias_cast, "format", None)),
                optional_arg=self.bias,
            )

            # Non Tensor Attributes (no need to quantize)
            normalized_shape = g.get_attr("normalized_shape")
            eps = g.get_attr("eps")

            args = ((_input_dq), normalized_shape, _weight_dq, _bias_dq, eps)
            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.layer_norm,
                args,
                name="ln",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class RMSNorm(DmxModule, torch.nn.RMSNorm):
    r"""
    An extension of RMSNorm layer to support DmxModule configurations.
    This module performs RMS-based layer normalization on the input tensor.
    The layer normalization is characterized by the `hidden_size` and an optional `eps` value for numerical stability.

    Args:
        dim (int): The size of the hidden layer (number of hidden units).
        eps (float, optional): A small constant added to the denominator for numerical stability. Defaults to 1e-6.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the RMS layer normalization.
    """

    def __init__(
        self,
        normalized_shape: int,
        eps: float = 1e-6,
    ) -> None:
        super().__init__(normalized_shape, eps=eps)
        self.functional_forward = F.rms_norm

    def _forward(self, _input: Tensor) -> Tensor:
        _output = self.approx_forward(
            (_input,), self.normalized_shape, self._weight, self.eps
        )
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        r"""
        Creates a new RMSNorm object (DmxModule) from a given PyTorch RMSNorm layer.

        Args:
            raw (torch.nn.Module): A PyTorch RMSNorm layer to be converted.

        Returns:
            DmxModule: A RMSNorm object that has the same configuration as the input PyTorch RMSNorm layer.
        """
        initial_dmx = cls(
            normalized_shape=raw.weight.shape[0],
            eps=raw.variance_epsilon if hasattr(raw, "variance_epsilon") else raw.eps,
        )
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            _weight_dq = g.get_attr(
                "weight", "weight_cast", repr(self.weight_cast.format)
            )

            # Non Tensor Attributes (no need to quantize)
            normalized_shape = g.get_attr("normalized_shape")
            eps = g.get_attr("eps")

            args = ((_input_dq), normalized_shape, _weight_dq, eps)
            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.rms_norm,
                args,
                name="RMSNorm",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class BatchNorm2d(DmxModule, torch.nn.BatchNorm2d):
    r"""
    An extension of PyTorch's BatchNorm2d layer to support DmxModule configurations.
    This module applies batch normalization over a 4D input tensor, suitable for use with 2D convolutional layers.
    The module is parameterized by the number of features, epsilon value for numerical stability, momentum for the running mean and variance, and options to use affine transformation and track running statistics.

    Args:
        num_features (int): Number of channels in the input tensor.
        eps (float, optional): A small constant added to the denominator for numerical stability. Defaults to 1e-05.
        momentum (float, optional): The momentum value for the running mean and running variance computation. Defaults to 0.1.
        affine (bool, optional): Whether to include learnable affine parameters for this layer. Defaults to True.
        track_running_stats (bool, optional): Whether to track the running mean and variance during training. Defaults to True.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the batch normalization.
    """

    def __init__(
        self,
        num_features: int,
        eps: float = 1e-05,
        momentum: float = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
    ) -> None:
        super().__init__(
            num_features,
            eps=eps,
            momentum=momentum,
            affine=affine,
            track_running_stats=track_running_stats,
        )

    def _forward(self, _input: Tensor) -> Tensor:
        self._check_input_dim(_input)
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum
        if self.training and self.track_running_stats:
            if self.num_batches_tracked is not None:
                self.num_batches_tracked = self.num_batches_tracked + 1
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum
        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is None)
        _output = F.batch_norm(
            _input,
            (
                self.running_mean
                if not self.training or self.track_running_stats
                else None
            ),
            self.running_var if not self.training or self.track_running_stats else None,
            self._weight,
            self._bias,
            bn_training,
            exponential_average_factor,
            self.eps,
        )
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        r"""
        Creates a new BatchNorm2d object (DmxModule) from a given PyTorch BatchNorm2d layer.

        Args:
            raw (torch.nn.Module): A PyTorch BatchNorm2d layer to be converted.

        Returns:
            DmxModule: A BatchNorm2d object that has the same configuration as the input PyTorch BatchNorm2d layer.
        """
        initial_dmx = cls(
            raw.num_features,
            eps=raw.eps,
            momentum=raw.momentum,
            affine=raw.affine,
            track_running_stats=raw.track_running_stats,
        )
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx


class GroupNorm(DmxModule, torch.nn.GroupNorm):
    r"""
    An extension of PyTorch's GroupNorm layer to support DmxModule configurations.
    This module applies group normalization over an input tensor, suitable for use with various types of layers.
    The module is parameterized by the number of groups, number of channels, epsilon value for numerical stability, and an option to use affine transformation.

    Args:
        num_groups (int): Number of groups to separate the channels into.
        channels (int): Number of channels in the input tensor.
        eps (float, optional): A small constant added to the denominator for numerical stability. Defaults to 1e-5.
        affine (bool, optional): Whether to include learnable affine parameters for this layer. Defaults to True.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the group normalization.
    """

    def __init__(
        self,
        num_groups: int,
        num_channels: int,
        eps: float = 1e-5,
        affine: bool = True,
    ) -> None:
        super().__init__(num_groups, num_channels, eps=eps, affine=affine)

    def _forward(self, input: Tensor) -> Tensor:
        _weight = self._weight
        _bias = self._bias
        _input = input.to(_weight.dtype) if _weight is not None else input

        _output = F.group_norm(_input, self.num_groups, _weight, _bias, self.eps)

        return _output

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            _weight_dq = g.get_attr(
                "weight", "weight_cast", repr(self.weight_cast.format)
            )
            _bias_dq = g.get_attr(
                "bias",
                "bias_cast",
                repr(getattr(self.bias_cast, "format", None)),
                optional_arg=self.bias,
            )

            args = (_input_dq, self.num_groups, _weight_dq, _bias_dq, self.eps)

            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.group_norm,
                args,
                name="GroupNorm",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class Dropout(DmxModule, torch.nn.Dropout):
    r"""
    An extension of PyTorch's Dropout layer to support DmxModule configurations.
    This module applies the dropout operation over the input tensor.

    Args:
        p (float, optional): The probability of an element to be zeroed. Defaults to 0.5.
        inplace (bool, optional): If set to ``True``, will do this operation in-place. Defaults to False.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the dropout layer.
    """

    def __init__(self, p: float = 0.5, inplace: bool = False) -> None:
        super().__init__(p=p, inplace=inplace)

    def _forward(self, _input: Tensor, *args, **kwargs) -> Tensor:
        _output = self.approx_forward((_input,))
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        r"""
        Creates a new Dropout object (DmxModule) from a given PyTorch Dropout layer.

        Args:
            raw (torch.nn.Module): A PyTorch Dropout layer to be converted.

        Returns:
            DmxModule: A Dropout object that has the same configuration as the input PyTorch Dropout layer.
        """
        initial_dmx = cls(p=raw.p, inplace=raw.inplace)
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            p = g.get_attr("p")
            training = g.get_attr("training")
            inplace = g.get_attr("inplace")

            args = (_input_dq, p, training, inplace)
            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.dropout,
                args,
                name="Dropout",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class ReLU(DmxModule, torch.nn.ReLU):
    r"""
    An extension of PyTorch's ReLU layer to support DmxModule configurations.
    This module applies the Rectified Linear Unit (ReLU) function element-wise on the input tensor.

    Args:
        inplace (bool, optional): If set to ``True``, will do this operation in-place. Defaults to False.

    Methods:
        _forward (_input: Tensor, inplace: bool = False) -> Tensor: Computes the forward pass of the ReLU layer.
    """

    def __init__(self, inplace: bool = False) -> None:
        super().__init__(inplace=inplace)

    def _forward(self, _input: Tensor, inplace: bool = False) -> Tensor:
        _output = self.approx_forward((_input,))
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        r"""
        Creates a new ReLU object (DmxModule) from a given PyTorch ReLU layer.

        Args:
            raw (torch.nn.Module): A PyTorch ReLU layer to be converted.

        Returns:
            DmxModule: A ReLU object that has the same configuration as the input PyTorch ReLU layer.
        """
        initial_dmx = cls(inplace=raw.inplace)
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )

            args = (_input_dq,)
            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.relu,
                args,
                name="ReLU",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class ReLU6(DmxModule, torch.nn.ReLU6):
    r"""
    An extension of PyTorch's ReLU6 layer to support DmxModule configurations.
    This module applies the Rectified Linear Unit 6 (ReLU6) function element-wise on the input tensor.

    Args:
        inplace (bool, optional): If set to ``True``, will do this operation in-place. Defaults to False.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the ReLU6 layer.
    """

    def __init__(self, inplace: bool = False) -> None:
        super().__init__(inplace=inplace)

    def _forward(self, _input: Tensor) -> Tensor:
        _output = self.approx_forward((_input,))
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        r"""
        Creates a new ReLU6 object (DmxModule) from a given PyTorch ReLU6 layer.

        Args:
            raw (torch.nn.Module): A PyTorch ReLU6 layer to be converted.
        Returns:
            DmxModule: A ReLU6 object that has the same configuration as the input PyTorch ReLU6 layer.
        """
        initial_dmx = cls(inplace=raw.inplace)
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )

            args = (_input_dq,)
            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.relu6,
                args,
                name="relu6",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class SiLU(DmxModule, torch.nn.SiLU):
    r"""
    An extension of PyTorch's SiLU (Sigmoid Linear Unit) layer to support DmxModule configurations.
    This module applies the SiLU function element-wise on the input tensor.

    Args:
        inplace (bool, optional): If set to ``True``, will do this operation in-place. Defaults to False.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the SiLU layer.
    """

    def __init__(self, inplace: bool = False) -> None:
        super().__init__(inplace=inplace)

    def _forward(self, _input: Tensor) -> Tensor:
        _output = self.approx_forward((_input,))
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        r"""
        Creates a new SiLU object (DmxModule) from a given PyTorch SiLU layer.

        Args:
            raw (torch.nn.Module): A PyTorch SiLU layer to be converted.

        Returns:
            DmxModule: A SiLU object that has the same configuration as the input PyTorch SiLU layer.
        """
        initial_dmx = cls(inplace=raw.inplace)
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )

            inplace = g.get_attr("inplace")
            args = (_input_dq, inplace)
            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.silu,
                args,
                name="SiLU",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class Tanh(DmxModule, torch.nn.Tanh):
    r"""
    An extension of PyTorch's Tanh (Hyperbolic Tangent) layer to support DmxModule configurations.
    This module applies the tanh function element-wise on the input tensor.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the Tanh layer.
    """

    def __init__(self) -> None:
        super().__init__()

    def _forward(self, _input: Tensor) -> Tensor:
        _output = self.approx_forward((_input,))
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        r"""
        Creates a new Tanh object (DmxModule) from a given PyTorch Tanh layer.

        Args:
            raw (torch.nn.Module): A PyTorch Tanh layer to be converted.

        Returns:
            DmxModule: A Tanh object that has the same configuration as the input PyTorch Tanh layer.
        """
        initial_dmx = cls()
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            args = (_input_dq,)

            _output_dq = g.create_node(
                "call_function",
                torch.nn.functional.tanh,
                args,
                name="tanh",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class GELUBase(DmxModule):
    r"""
    A generalized base class to support various GELUActivation configurations.
    This module applies the specified GELUActivation function element-wise on the input tensor.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the GELU layer.
    """

    def __init__(self, activation_cls, *args, **kwargs) -> None:
        if activation_cls not in self.__class__.__bases__:
            self.__class__.__bases__ += (activation_cls,)
        super().__init__(*args, **kwargs)
        self.activation_cls = activation_cls

    def _forward(self, _input: Tensor) -> Tensor:
        _output = self.approx_forward((_input,))
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        r"""
        Creates a new GELU object (DmxModule) from a given Transformers layer.

        Args:
            raw (torch.nn.Module): A Transformers GELUActivation layer to be converted.

        Returns:
            DmxModule: A GELU object that has the same configuration as the input Transformers GELUActivation layer.
        """
        initial_dmx = cls()
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            args = (_input_dq,)

            _output_dq = g.create_node(
                "call_function",
                self.functional_forward,
                args,
                name="gelu",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class GELU(GELUBase):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(torch.nn.GELU, *args, **kwargs)

    def functional_forward(self, x):
        return F.gelu(x)



#!/usr/bin/env python3
#
from .tracer import QuantTracer
from .transformer.input_output_transformer import InputOutputTransformer
from .transformer.qdq_transformer import QdQTransformer
from .transformer.configuration_transformer import ConfigurationTransformer
from .transformer.dmx_aware_transformer import DMXAwareTransformer
from .transformer.nodedict_transformer import NodeDictTransformer
from .transformer.record_input_interpreter import RecordInputInterpreter



from typing import Optional, Dict
from transformers import pipeline as hfpipeline
from dmx.compressor.fx.transform import substitute_transform
import transformers
from .model import DmxModelMixin, DmxConfig
import evaluate
from evaluate import evaluator
from datasets import load_dataset
from huggingface_hub import hf_hub_download
from .model import DmxModel, DmxConfig
from tqdm import tqdm
import torch
from transformers import AutoModelForCausalLM
from accelerate.utils import compute_module_sizes, find_tied_parameters
import re
from dmx.compressor import config_rules


def get_config_file(repo_name, revision, config_name):
    if config_name in ["BASELINE", "BASIC"]:
        return None
    try:
        return hf_hub_download(
            repo_id=repo_name, filename=f"configs/{config_name}.yaml", revision=revision
        )
    except Exception as e:
        print(f"Failed to download the file: {str(e)}")
        return None


def dmx_transform(pipe, dmx_config_name):
    config_file = get_config_file(pipe.model_name, pipe.revision, dmx_config_name)
    if config_file is not None:
        config = DmxConfig.from_yaml(config_file)
        pipe.model.configure(config)
    else:
        if dmx_config_name in ["BASELINE", "BASIC"]:
            # NOTE: assuming pipe.model is in BASELINE mode
            pipe.model.configure(None, *eval(f"config_rules.{dmx_config_name}"))
        else:
            raise RuntimeError(f"illegal dmx_config: {dmx_config_name}")


def prepare_dataset_and_column(
    dataset,
    column_name=None,
    dataset_version=None,
    dataset_split="test",
    seed=42,
    trust_remote_code=True,
):
    dataset_column_mapping = {
        "wikitext": "text",
        "ptb_text_only": "sentence",
        "lambada": "text",
        "EleutherAI/lambada_openai": "text",
        # Add more datasets and their respective column names here
    }

    if dataset == "squad":
        column_name = None
    elif not column_name and dataset in dataset_column_mapping:
        column_name = dataset_column_mapping[dataset]
    elif not column_name:
        raise ValueError(
            f"Column name not found for dataset '{dataset}'. Please provide the column_name."
        )

    dataset = load_dataset(
        dataset,
        dataset_version,
        split=dataset_split,
        trust_remote_code=trust_remote_code,
    )
    if dataset_split == "train":
        dataset = dataset.shuffle(seed=seed)

    return dataset, column_name


def do_forward_on(
    model,
    tokenizer,
    dataset,
    column_name=None,
    dataset_version=None,
    dataset_split="test",
    num_samples=None,
    seed=42,
):
    dataset, column_name = prepare_dataset_and_column(
        dataset, column_name, dataset_version, dataset_split, seed
    )

    encodings = tokenizer("\n\n".join(dataset[column_name]), return_tensors="pt")

    if hasattr(model.config, "max_position_embeddings"):
        max_seq_len = model.config.max_position_embeddings
    elif hasattr(model.config, "n_positions"):
        max_seq_len = model.config.n_positions
    else:
        max_seq_len = 2048

    stride = max_seq_len

    seq_len = encodings.input_ids.size(1)

    if num_samples:
        seq_len = min((num_samples - 1) * stride + max_seq_len, seq_len)
    else:
        seq_len = (seq_len // stride) * stride

    nlls = []
    prev_end_loc = 0
    model.eval()

    for begin_loc in tqdm(range(0, seq_len, stride)):
        end_loc = min(begin_loc + max_seq_len, seq_len)
        trg_len = end_loc - prev_end_loc
        input_ids = encodings.input_ids[:, begin_loc:end_loc].to(model.device)
        target_ids = input_ids.clone()
        target_ids[:, :-trg_len] = -100

        with torch.no_grad():
            outputs = model(input_ids, labels=target_ids)
            neg_log_likelihood = outputs.loss * trg_len

        nlls.append(neg_log_likelihood.to(model.device))

        prev_end_loc = end_loc
        if end_loc == seq_len:
            break

    loss = torch.stack(nlls).float().sum() / end_loc
    ppl = torch.exp(loss)

    return dict(
        loss=loss.item(),
        perplexity=ppl.item(),
    )


def eval_question_answering(
    model,
    tokenizer,
    dataset,
    metric=None,
    dataset_split="test",
    revision="main",
    column_name=None,
    dataset_version=None,
):
    dataset, column_name = prepare_dataset_and_column(
        dataset, column_name, dataset_version, dataset_split
    )
    task_evaluator = evaluator("question-answering")
    results = task_evaluator.compute(
        model_or_pipeline=model, tokenizer=tokenizer, data=dataset
    )
    return results


def eval_text_generation(
    model,
    tokenizer,
    dataset,
    metric,
    dataset_split="test",
    revision="main",
    column_name=None,
    dataset_version=None,
):
    dataset, column_name = prepare_dataset_and_column(
        dataset, column_name, dataset_version, dataset_split
    )

    metric = evaluate.load(metric, module_type="metric")

    results = metric.compute(
        model=model,
        tokenizer=tokenizer,
        revision=revision,
        references=dataset[column_name],
    )
    return results


def pipe_eval(
    model,
    tokenizer,
    dataset,
    metric,
    task,
    revision,
    dataset_split="test",
    column_name=None,
    dataset_version=None,
):
    task_eval_mapping = {
        "text-generation": eval_text_generation,
        "question-answering": eval_question_answering,
    }

    if task not in task_eval_mapping:
        raise ValueError(f"Unsupported task type '{task}'.")

    eval_function = task_eval_mapping[task]
    return eval_function(
        model,
        tokenizer,
        dataset,
        metric,
        dataset_split,
        revision,
        column_name,
        dataset_version,
    )


def contains_number(string):
    """
    This function checks whether a string contains a number or not.
    This is used to identify hidden layers in the model.
    """
    return bool(re.search(r"\d", string))


def get_modules(root: torch.nn.Module, prefix: str) -> Dict[str, torch.nn.Module]:
    """
    A function that recursively traverses the model from the given root module and
    returns a dictionary of submodules for device mapping. In accordance with the format
    of device_map = "auto", only submodules that are leaf nodes or hidden layers are
    included in the dictionary; submodules of hidden layers are ignored.

    Args:
        root (torch.nn.Module): model/module to traverse
        prefix (str): prefix for the submodule names

    Returns:
        Dict[str, torch.nn.Module]: dictionary of submodules
    """
    modules = {}
    for name, module in root.named_children():
        if len(list(module.children())) == 0 or contains_number(name):
            modules[prefix + name] = module
        else:
            submodules = get_modules(module, prefix + name + ".")
            modules.update(submodules)
    return modules


def balanced_device_map(model: str, revision: Optional[str] = "main") -> Dict[str, int]:
    """
    A function that computes a custom device map for the given model that distributes model weights
    evenly across all devices. Enable with device_map = "balanced" when calling pipeline.

    Args:
        model (str): model name on huggingface
        revision (str): revision of the model on huggingface

    Returns:
        Dict[str, int]: dictionary of the device map
    """
    model = AutoModelForCausalLM.from_pretrained(
        model, revision=revision, device_map="meta"
    )
    module_sizes = compute_module_sizes(model)
    tied_parameters = find_tied_parameters(model)
    modules = get_modules(model, "")

    params = {}
    for k, v in modules.items():
        params[k] = module_sizes[k]

    device_map = {}
    total_params = sum(params.values())
    num_devices = torch.cuda.device_count()

    cur_device = 0
    params_on_cur_device = 0
    accumulated_params = 0
    average_params = total_params // num_devices

    # place all embedding modules together on the first device
    for name, module in modules.items():
        if isinstance(module, torch.nn.Embedding):
            device_map[name] = cur_device
            params_on_cur_device += params[name]
            continue

    # distribute remaining modules across all devices
    for name, module in modules.items():
        if isinstance(module, torch.nn.Embedding):
            continue
        if params_on_cur_device > average_params or (
            params_on_cur_device != 0
            and params_on_cur_device + params[name] > average_params * 1.2
        ):
            cur_device = min(cur_device + 1, num_devices - 1)
            accumulated_params += params_on_cur_device
            params_on_cur_device = 0
            average_params = (total_params - accumulated_params) // (
                num_devices - cur_device
            )
        device_map[name] = cur_device
        params_on_cur_device += params[name]

    for pair in tied_parameters:
        device_map[pair[0][: -len(".weight")]] = device_map[pair[1][: -len(".weight")]]

    return device_map


def pipeline(
    *args,
    dmx_config: Optional[str] = None,
    trust_remote_code: bool = True,
    device_map: Optional[str] = "auto",
    **kwargs,
):
    if device_map == "balanced":
        device_map = balanced_device_map(
            kwargs.get("model"), kwargs.get("revision", "main")
        )
    kwargs.update(
        {
            "trust_remote_code": trust_remote_code,
            "device_map": device_map,
        }
    )
    pipe = hfpipeline(*args, **kwargs)
    pipe.task = kwargs.get("task")
    pipe.model_name = kwargs.get("model")
    pipe.revision = kwargs.get("revision", "main")
    pipe.model = DmxModel.from_torch(pipe.model)
    pipe.evaluate = lambda metric, dataset, column_name=None, dataset_version=None, dataset_split="test": pipe_eval(
        pipe.model,
        pipe.tokenizer,
        dataset,
        metric,
        pipe.task,
        pipe.revision,
        dataset_split,
        column_name,
        dataset_version,
    )

    pipe.do_forward_on = lambda dataset, column_name=None, dataset_version=None, dataset_split="test", num_samples=None, seed=42: do_forward_on(
        pipe.model,
        pipe.tokenizer,
        dataset,
        column_name,
        dataset_version,
        dataset_split,
        num_samples,
        seed,
    )

    dmx_transform(pipe, dmx_config)

    return pipe


class DmxPreTrainedModel(transformers.modeling_utils.PreTrainedModel, DmxModelMixin):
    @classmethod
    def from_pretrained(cls, *args, **kwargs):
        _model = super().from_pretrained(*args, **kwargs)
        _model.base_model = substitute_transform(_model.base_model, hf=True)
        return _model



from typing import Optional
import torch
import numpy as np
import transformers
import warnings
from contextlib import contextmanager


def _count_conv_flops(
    layer: torch.nn.modules.conv._ConvNd,
    _input: torch.Tensor,
    _output: torch.Tensor,
) -> int:
    r"""
    Taken from https://github.com/sovrasov/flops-counter.pytorch/blob/master/ptflops/pytorch_ops.py
    """
    batch_size = _input.shape[0]
    output_dims = list(_output.shape[2:])
    kernel_dims = list(layer.kernel_size)
    in_channels = layer.in_channels
    out_channels = layer.out_channels
    groups = layer.groups
    filters_per_channel = out_channels // groups
    conv_per_position_flops = (
        int(np.prod(kernel_dims)) * in_channels * filters_per_channel
    )
    active_elements_count = batch_size * int(np.prod(output_dims))
    return conv_per_position_flops * active_elements_count


def _count_linear_flops(
    layer: torch.nn.Linear,
    _input: torch.Tensor,
    _output: torch.Tensor,
) -> int:
    return int(np.prod(list(_input.shape))) * layer.out_features


def _count_hf_transformers_conv1d_flops(
    layer: transformers.pytorch_utils.Conv1D,
    _input: torch.Tensor,
    _output: torch.Tensor,
) -> int:
    return int(np.prod(list(_input.shape))) * layer.nf


class PerformanceProxyMixin:
    flop_counter: Optional[int] = None
    flop_counter_enabled: bool = False
    last_input_shape: Optional[torch.Size] = None
    last_output_shape: Optional[torch.Size] = None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def zero_flop_counter(self) -> None:
        self.flop_counter = 0

    def enable_flop_counter(self, state: bool = True) -> None:
        self.flop_counter_enabled = state
        if self.flop_counter_enabled and self.flop_counter is None:
            self.zero_flop_counter()

    def count_flops(
        self,
        _input: torch.Tensor,
        _output: torch.Tensor,
    ) -> None:
        if self.flop_counter is not None:
            self.last_input_shape = _input.shape
            self.last_output_shape = _output.shape
            if isinstance(self, torch.nn.modules.conv._ConvNd):
                self.flop_counter += _count_conv_flops(self, _input, _output)
            elif isinstance(self, torch.nn.Linear):
                self.flop_counter += _count_linear_flops(self, _input, _output)
            elif isinstance(self, transformers.pytorch_utils.Conv1D):
                self.flop_counter += _count_hf_transformers_conv1d_flops(
                    self, _input, _output
                )
            else:
                self.flop_counter = None
                warnings.warn(
                    f"cannot count flops of this kind of layer: {type(self)}",
                    RuntimeWarning,
                )

    def _has_weight(self) -> bool:
        return "weight" in [n for n, _ in self.named_parameters()]

    @property
    def weight_elem_count(self) -> Optional[float]:
        if self._has_weight():
            _n = float(self.weight.nelement())
            if self.weight_sparsifier is not None:
                _n *= (
                    self.weight_sparsifier.density
                )  # not considering compression scheme and the overhead coming with different schemes
            return _n
        else:
            return None

    @property
    def weight_size_in_bytes(self) -> Optional[float]:
        if self._has_weight():
            if self.weight_cast is not None:
                bytes_per_elem = (
                    self.weight.element_size()
                    if self.weight_cast.format.bytes_per_elem is None
                    else self.weight_cast.format.bytes_per_elem
                )
            return bytes_per_elem * self.weight_elem_count
        else:
            return None

    @property
    def flops(self) -> Optional[float]:
        _flops = self.flop_counter
        if (
            _flops is not None
            and self._has_weight()
            and self.weight_sparsifier is not None
        ):
            _flops *= self.weight_sparsifier.density
        return _flops

    @property
    def bops(self) -> Optional[float]:
        _bops = self.flops
        if _bops is not None and self._has_weight():
            _bops *= self.input_precision * self.weight_precision
        return _bops

    @contextmanager
    def counting_flops(self, zero: bool = True) -> None:
        self.enable_flop_counter(True)
        if zero:
            self.zero_flop_counter()
        yield self
        self.enable_flop_counter(False)



import torch
import torch.nn as nn

from dmx.compressor.fx.tracer import HFQuantTracer, hf_symbolic_trace
from ..fx import QuantTracer, InputOutputTransformer, QdQTransformer
from dmx.compressor.fx import (
    ConfigurationTransformer,
    DMXAwareTransformer,
    RecordInputInterpreter,
)
from dmx.compressor.fx.transformer.utils import dmx_aware_mapping
from torch.fx import GraphModule
from typing import Any, Dict, List, Optional, Union
from inspect import signature


def prepare_tracing_inputs(_model, args, kwargs):
    # remove kwargs with value None
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    # boolean inputs will affect tracing and need to be set as concrete args
    bool_inputs = {k: v for k, v in kwargs.items() if isinstance(v, bool)}
    kwargs = {k: v for k, v in kwargs.items() if not isinstance(v, bool)}

    if hasattr(_model, "old_forward"):
        input_names = (
            signature(_model.old_forward).bind(*args, **kwargs).arguments.keys()
        )
    else:
        input_names = signature(_model.forward).bind(*args, **kwargs).arguments.keys()
    dummy_inputs = {}
    i = 0
    for k in input_names:
        if k not in kwargs:
            dummy_inputs[k] = args[i]
            i += 1
        else:
            dummy_inputs[k] = kwargs[k]
    return input_names, bool_inputs, dummy_inputs


def substitute_transform(
    root: torch.nn.Module,
    concrete_args: Optional[Dict[str, Any]] = None,
    input_names: Optional[List[str]] = None,
    dummy_inputs: Optional[Dict[str, Any]] = None,
    additional_mappings: Optional[Dict[str, Any]] = None,
):
    """
    A function that transforms the model by substituting torch.nn.modules and activation functions to dmx.nn.modules.

    Args:
        root (torch.nn.Module): model/module to transform
        concrete_args (Dict[str,Any], optional): concrete arguments to be used for tracer. Defaults to None.
        hf (bool, optional): True if root is a huggingface model. Defaults to False

    Returns:
        transformed model
    """
    mod_type = type(root).__module__ + "." + type(root).__name__
    if mod_type in dmx_aware_mapping:
        transformed = dmx_aware_mapping[mod_type].from_raw(root)
        return transformed

    gm, tracer = hf_symbolic_trace(
        root,
        input_names,
        concrete_args=concrete_args,
        dummy_inputs=dummy_inputs,
    )

    gi = RecordInputInterpreter(gm)
    gi.run(*list(dummy_inputs.values()))
    transformer = DMXAwareTransformer(
        gm,
        tracer.node_name_to_scope,
        root._gm if root.transformed else None,
        gi.nodeInputs,
    )
    if additional_mappings:
        for target, dmx_module in additional_mappings.items():
            new_target = target.replace("torch.ops.", "")
            transformer.add_dmx_aware_functional_mapping(new_target, dmx_module)
    transformed = transformer.transform()

    return transformed


def qDq_transform(
    root: torch.fx.GraphModule,
):
    """
    A function that transforms the model by substituting CastTos with Q/dQ ops

    Args:
        root (torch.nn.Module): model/module to transform

    Returns:
        transformed model
    """
    # import compiler so q/dq ops are registered

    transformer = QdQTransformer(root)
    transformed = transformer.transform()
    transformed.recompile()
    return transformed


make_compiler_graph = qDq_transform


def cast_input_output_transform(
    root: torch.nn.Module,
    tracer: Union[QuantTracer, HFQuantTracer] = QuantTracer(),
    concrete_args: Optional[Dict[str, Any]] = None,
    cfg: Optional[str] = None,
) -> nn.Module:
    """
    A function that transforms the module by adding additional ops, which includes:
    - casting
    - approximator
    - sparsifier
    An optional config file can be passed to specify the formats for the additional ops, dummy formats would be used otherwise.

    Args:
        root (torch.nn.Module): model/module to transform
        tracer (Union[QuantTracer, HFQuantTracer], optional): tracer used for tracing the root. Defaults to QuantTracer.
        concrete_args (Dict[str,Any], optional): concrete arguments to be used for tracer. Defaults to None.
        cfg (Optional[str]): config file for setting the added ops formats. Defaults to None.

    Returns:
        transformed model
    """
    graph = tracer.trace(root, concrete_args)
    name = (
        root.__class__.__name__ if isinstance(root, torch.nn.Module) else root.__name__
    )
    gm = GraphModule(tracer.root, graph, name)
    transformer = InputOutputTransformer(gm, tracer.node_name_to_scope, cfg)
    transformed = transformer.transform()
    transformed.scopeDict = transformer.scopeDict
    return transformed


def configure_transform(gm: torch.fx.GraphModule, scopeDict: dict, cfg: str):
    """
    A function that changes the format of the ops according to the cfg file
    Note:
        Configure_transform will only change existing ops and will not add any additional ops.
        Hence it is recommened to pass in a cfg file for cast_input_output_transform to make sure all necessary ops are added.

    Args:
        gm (torch.fx.GraphModule): Graphmodule to apply changes on
        scopeDict (dict): Dictionary that maps node name to scope
        cfg (str): config file for setting the added ops formats.

    Returns:
        Graphmodule with updated formats
    """
    transformer = ConfigurationTransformer(gm, scopeDict, cfg)
    transformer.transform()
    return gm



import pytest
import torch
import torch.nn as nn
import numpy as np
from dmx.compressor import DmxModel
from dmx.compressor.modeling import nn as dmxnn
from skopt.space import Categorical

from dmx.compressor.advanced_recipe import (
    DmxApproximationFunctionTuningHyperparams,
    DmxApproximationFunctionTuningRecipe,
)

from dmx.compressor.advanced_recipe import DmxModuleSmoothQuantHyperparams


class TestNetwork(nn.Module):
    def __init__(self):
        super().__init__()
        self.lin1 = nn.Linear(40,50)
        self.lnorm1 = torch.nn.LayerNorm(50)
        self.lin2 = nn.Linear(50,60)
        self.lnorm2 = torch.nn.LayerNorm(60)
    def forward(self,x):
        x = self.lnorm1(self.lin1(x))
        x = torch.nn.functional.relu(x)
        return self.lnorm2(self.lin2(x))

RANDOM_SEED = 0

torch.manual_seed(RANDOM_SEED)

def test_simd_tuning(
):
    model = TestNetwork()
    model = DmxModel.from_torch(model)
    model.to_basic_mode()
    model(torch.randn(10,40))
    
    def hp_gen(_model):
        return {
            _m: DmxApproximationFunctionTuningHyperparams(
                search_space=[
                    Categorical(
                        np.logspace(start=-5, stop=1, num=7, base=2), 
                        name="norm",
                    )
                ],
            )
            for _, _m in _model.named_dmx_modules()
            if isinstance(_m, (nn.LayerNorm,))
        }
    with DmxApproximationFunctionTuningRecipe(hp_gen).applied_to(model):
        model(torch.randn(10,40))
    
    assert model._gm.lnorm1.approximator.function is not \
        model._gm.lnorm2.approximator.function
    
    assert 'norm' in model._gm.lnorm1.approximator.function.extra_params
    assert 'norm' in model._gm.lnorm2.approximator.function.extra_params



from abc import ABC
from typing import Any, Optional
from parse import parse
from bidict import bidict
import torch

try:
    from ..quant import fixed_point_quantize, block_quantize, float_quantize
except ImportError as error:
    print("Error importing Block Quantize CUDA kernels")
from .onnx import BFPTypeEnum

try:
    from numerics import (
        determine_sbfp_scaler_exponent_bias_from_tensor_values,
    )

    NUMERICS_UTILS_AVAILABLE = True
except ModuleNotFoundError:
    NUMERICS_UTILS_AVAILABLE = False


ROUNDING_MODE = bidict(
    {
        "U": "up",
        "D": "down",
        "N": "nearest",
        "S": "stochastic",
    }
)


class Format(ABC):
    r"""
    This is an abstract class of tensor numerical format.
    Child classes to implement `cast()` and `from_shorthand()` method.
    """

    blocked: bool
    bfp_id: Optional[int] = None

    def __init__(self):
        pass

    def __str__(self) -> str:
        raise NotImplementedError

    def cast(self, *input: Any):
        raise NotImplementedError

    @property
    def bytes_per_elem(self) -> Optional[float]:
        raise NotImplementedError

    @staticmethod
    def from_shorthand(sh: str):
        if sh.startswith("SAME"):
            return Same.from_shorthand(sh)
        elif sh.startswith("XP"):
            return FixedPoint.from_shorthand(sh)
        elif sh.startswith("FP"):
            return FloatingPoint.from_shorthand(sh)
        elif sh.startswith("BFP"):
            return BlockFloatingPoint.from_shorthand(sh)
        elif sh.startswith("SBFP"):
            return ScaledBlockFloatingPoint.from_shorthand(sh)
        elif sh.startswith("MXFP"):
            return MXFP.from_shorthand(sh)
        elif sh.startswith("MXINT"):
            return MXINT.from_shorthand(sh)
        else:
            raise ValueError(f"unrecognized format shorthand: {sh}")

    @property
    def bit_precision(self) -> Optional[float]:
        raise NotImplementedError


class Same(Format):
    r"""
    This is a dummy numerical format whose `cast()` does not do anything but passing same input through.
    """

    blocked = False

    def __init__(self):
        super().__init__()

    def cast(self, x, *args):
        return x.clone()

    @property
    def bytes_per_elem(self) -> None:
        return None

    @classmethod
    def from_shorthand(cls, sh: str):
        return cls()

    @property
    def bit_precision(self) -> Optional[float]:
        return None

    def __str__(self) -> str:
        return "Dummy numerical format: no casting"

    def __repr__(self) -> str:
        return "SAME"


class FixedPoint(Format):
    r"""
    This is a fixed point format simulated in FP32, using QPyTorch.
    """

    blocked = False

    def __init__(
        self, precision, fraction, clamp=True, symmetric=True, rounding="nearest"
    ):
        super().__init__()
        # check validity of format configuration
        assert (
            1 <= precision <= 24
        ), f"highest integer precision simulated by FP32 is 25, got {precision}"
        # [TODO] check fraction validity, considering 8-bit exponent of FP32

        self.precision = precision
        self.fraction = fraction
        self.clamp = clamp
        self.symmetric = symmetric
        self.rounding = rounding

    def cast(self, x, *args):
        return fixed_point_quantize(
            x,
            wl=self.precision,
            fl=self.fraction,
            clamp=self.clamp,
            symmetric=self.symmetric,
            rounding=self.rounding,
        )

    @property
    def bytes_per_elem(self) -> float:
        return self.precision / 8.0

    @classmethod
    def from_shorthand(cls, sh: str):
        conf = parse(
            "XP[{precision:d},{fraction:d}]({clamp:w}{symmetric:w}{rounding:w})", sh
        )
        return cls(
            precision=conf["precision"],
            fraction=conf["fraction"],
            clamp=conf["clamp"] == "C",
            symmetric=conf["symmetric"] == "S",
            rounding=ROUNDING_MODE[conf["rounding"]],
        )

    @property
    def bit_precision(self) -> float:
        return float(self.precision)

    def __str__(self) -> str:
        return f"Simulated fixed point format: precision bits = {self.precision}, fraction bits = {self.fraction}, \ncasting behavior: symmetric = {self.symmetric}, clamp = {self.clamp}, rounding = {self.rounding}"

    def __repr__(self) -> str:
        return f"XP[{self.precision},{'0' if self.fraction==0 else f'{self.fraction:+d}'}]({'C' if self.clamp else '_'}{'S' if self.symmetric else '_'}{ROUNDING_MODE.inverse[self.rounding]})"


class FloatingPoint(Format):
    r"""
    This is a floating point format simulated in FP32, using QPyTorch.
    """

    blocked = False

    def __init__(
        self,
        mantissa=23,
        exponent=8,
        bias=None,
        flush_subnormal=True,
        unsigned=False,
        rounding="nearest",
    ):
        super().__init__()
        # check validity of format configuration
        assert (
            0 <= mantissa <= 23
        ), f"number of mantisa bits simulatable by FP32 is between 0 and 23, got{mantissa}"
        assert (
            0 < exponent <= 8
        ), f"number of exponent bits simulatable by FP32 is between 1 and 8, got {exponent}"
        _bias_min = 127 if exponent == 8 else -128 + 2**exponent
        assert (
            _bias_min <= bias <= 127
        ), f"exponent bias simulatable by FP32 for {exponent}-bit exponent is constrained between {_bias_min} and 127, got {bias}"

        self.mantissa = mantissa
        self.exponent = exponent
        self.bias = bias if bias is not None else 2 ** (exponent - 1) - 1
        self.flush_subnormal = flush_subnormal
        self.unsigned = unsigned
        self.rounding = rounding

    def cast(self, x, *args):
        x = (
            x
            if (x.dtype == torch.float32 and repr(self) == "FP[1|8|23,127](_N)")
            or (x.dtype == torch.float16 and repr(self) == "FP[1|5|10,15](_N)")
            else float_quantize(
                x.float(),
                man=self.mantissa,
                exp=self.exponent,
                bias=self.bias,
                flush_subnormal=self.flush_subnormal,
                rounding=self.rounding,
            )
        )
        return x.abs() if self.unsigned else x

    @property
    def largest_representable_power_of_two(self):
        return 2 ** (2 ** (self.exponent - 1))

    @property
    def bytes_per_elem(self) -> float:
        return (self.mantissa + self.exponent + 1) / 8.0

    @classmethod
    def from_shorthand(cls, sh: str):
        conf = parse(
            "FP[{sign:d}|{exponent:d}|{mantissa:d},{bias:d}]({flush_subnormal:w}{rounding:l})",
            sh,
        )
        return cls(
            mantissa=conf["mantissa"],
            exponent=conf["exponent"],
            bias=conf["bias"],
            flush_subnormal=conf["flush_subnormal"] == "F",
            unsigned=conf["sign"] == 0,
            rounding=ROUNDING_MODE[conf["rounding"]],
        )

    @property
    def bit_precision(self) -> float:
        return float(
            self.mantissa + self.exponent
            if self.unsigned
            else 1 + self.mantissa + self.exponent
        )

    def __str__(self) -> str:
        return f"Simulated floating point format: mantissa bits = {self.mantissa}, exponent bits = {self.exponent}, exponent bias = {self.bias}, unsigned = {self.unsigned}, \ncasting behavior: flush subnormal = {self.flush_subnormal}, rounding = {self.rounding}"

    def __repr__(self) -> str:
        return f"FP[{'0' if self.unsigned else '1'}|{self.exponent}|{self.mantissa},{self.bias}]({'F' if self.flush_subnormal else '_'}{ROUNDING_MODE.inverse[self.rounding]})"


class BlockFloatingPoint(Format):
    r"""
    This is a block floating point format simulated in FP32, using QPyTorch.
    """

    blocked = True

    def __init__(
        self,
        precision=8,
        block_size=64,
        symmetric=True,
        rounding="nearest",
    ):
        super().__init__()
        # check validity of format configuration
        assert (
            2 <= precision <= 25
        ), f"highest integer precision simulated by FP32 is 25, got {precision}"
        assert block_size > 0, f"block size has to be positive, got {block_size}"

        self.precision = precision
        self.block_size = block_size
        self.symmetric = symmetric
        self.rounding = rounding

    @property
    def bfp_id(self):
        name = f"DMX_BFP_{self.precision+8}{'' if self.symmetric else 'A'}_{self.block_size}"
        return BFPTypeEnum[name].value

    def cast(self, x: torch.Tensor, block_dim: int):
        # input of Linear: [B, ..., Cin], dim=-1
        # weight of Linear: [Cout, Cin], dim=-1
        # input of Conv1D: [B, Cin, L], dim=1
        # weight of Conv1D: [Cout, Cin//G, K], dim=1
        # input of Conv2D: [B, Cin, H, W], dim=1
        # weight of Conv2D: [Cout, Cin//G, K, K], dim=1
        # TODO: modify qtorch kernels do the following at C++ level
        if self.block_size == 1:  # borrowing float_quantize for block_size==1
            _x = float_quantize(
                x.float(),
                man=self.precision - 2,  # 1 for sign and 1 for implicit bit
                exp=8,
                bias=127,
                flush_subnormal=False,
                rounding=self.rounding,
            )
        else:
            _x = x.float().transpose(block_dim, -1)  # dim swap
            xshape = _x.shape  # remember shape
            _xs = torch.split(
                _x.reshape((-1, xshape[-1])), self.block_size, dim=-1
            )  # slice to blocks
            _x = torch.cat(
                [
                    block_quantize(
                        block,
                        wl=self.precision,
                        dim=0,
                        symmetric=self.symmetric,
                        rounding=self.rounding,
                    )
                    for block in _xs
                ],
                dim=-1,
            )  # quantize
            _x = _x.reshape(xshape).transpose_(block_dim, -1)  # recover shape
        return _x

    @property
    def bytes_per_elem(self) -> float:
        return (self.precision + 8.0 / self.block_size) / 8.0

    @classmethod
    def from_shorthand(cls, sh: str):
        conf = parse(
            "BFP[{precision:d}|8]{{{block_size:d}}}({symmetric:w}{rounding:l})",
            sh,
        )

        return cls(
            precision=conf["precision"],
            block_size=conf["block_size"],
            symmetric=conf["symmetric"] == "S",
            rounding=ROUNDING_MODE[conf["rounding"]],
        )

    @property
    def bit_precision(self) -> float:
        return self.precision + 8.0 / self.block_size

    def __str__(self) -> str:
        return f"Simulated block floating point format: precision bits = {self.precision}, block size = {self.block_size}\ncasting behavior: symmetric = {self.symmetric}, rounding = {self.rounding}"

    def __repr__(self) -> str:
        return f"BFP[{self.precision}|8]{{{self.block_size}}}({'S' if self.symmetric else '_'}{ROUNDING_MODE.inverse[self.rounding]})"


class ScaledBlockFloatingPoint(Format):
    r"""
    This is a scaled block floating point tensor format.
    """

    blocked = True

    def __init__(
        self,
        block_format: FixedPoint,
        scaler_format: FloatingPoint,
        block_size=64,
    ):
        super().__init__()
        # check validity of format configuration
        assert isinstance(
            block_format, FixedPoint
        ), "block format needs to be fixed point"
        assert isinstance(
            scaler_format, FloatingPoint
        ), "scaler format needs to be floating point"
        assert block_format.fraction == 0, "block format needs to have zero fraction"
        assert block_format.symmetric, "block format needs to have symmetric range"
        assert block_size > 0, f"block size has to be positive, got {block_size}"

        self.block_format = block_format
        self.scaler_format = scaler_format
        self.block_size = block_size

        self.man_scaling = (
            2 ** (self.block_format.precision - 1) - 1
        )  # largest mantissa abs
        self.get_chunk_max = lambda chunk: torch.max(
            torch.abs(chunk), dim=-1, keepdim=True
        )[0]

        self.scaler_format_exponent_bias_determined = False

    def determine_scaler_exponent_bias_from(self, x: torch.Tensor) -> None:
        r"""
        Determines scaler_format.bias based on the value of the quantized tensor
        as a side-effect
        """
        if NUMERICS_UTILS_AVAILABLE:
            self.scaler_format.bias = (
                determine_sbfp_scaler_exponent_bias_from_tensor_values(x)
            )

    @property
    def bfp_id(self):
        name = f"DMX_SBFP_{self.block_format.precision+8}_{self.block_size}_{self.scaler_format.bias}"
        return BFPTypeEnum[name].value

    def cast(self, x: torch.Tensor, block_dim: int) -> torch.Tensor:
        if not self.scaler_format_exponent_bias_determined:
            self.determine_scaler_exponent_bias_from(x)
            self.scaler_format_exponent_bias_determined = True
        _x = x.float().transpose(block_dim, -1)  # dim swap
        xshape = _x.shape  # remember shape
        _xs = torch.split(
            _x.reshape((-1, xshape[-1])), self.block_size, dim=-1
        )  # slice to chunks
        _xms = [
            self.get_chunk_max(chunk) / self.man_scaling for chunk in _xs
        ]  # max of blocks in each chunk
        _x = torch.cat(
            [
                torch.where(
                    chunk_max > 0.0,
                    self.block_format.cast(chunk / chunk_max)
                    * self.scaler_format.cast(chunk_max),
                    chunk,
                )
                for chunk, chunk_max in zip(_xs, _xms)
            ],
            dim=-1,
        )  # quantize

        _x = _x.reshape(xshape).transpose_(block_dim, -1)  # recover shape
        return _x

    @property
    def bytes_per_elem(self) -> float:
        return (
            self.block_format.bytes_per_elem
            + self.scaler_format.bytes_per_elem / self.block_size
        )

    @classmethod
    def from_shorthand(cls, sh: str):
        conf = parse(
            "SBFP<{block_format_sh}><{scaler_format_sh}>{{{block_size:d}}}",
            sh,
        )
        return cls(
            block_format=FixedPoint.from_shorthand(conf["block_format_sh"]),
            scaler_format=FloatingPoint.from_shorthand(conf["scaler_format_sh"]),
            block_size=conf["block_size"],
        )

    @property
    def bit_precision(self) -> float:
        return (
            self.block_format.bit_precision
            + self.scaler_format.bit_precision / self.block_size
        )

    def __str__(self) -> str:
        return f"Simulated scaled block floating point format: block format = {self.block_format}, scaler format = {self.scaler_format},\n block size = {self.block_size}"

    def __repr__(self) -> str:
        return f"SBFP<{repr(self.block_format)}><{repr(self.scaler_format)}>{{{self.block_size}}}"


class MXFP(Format):
    r"""
    This is a MXFP tensor format.
    """

    blocked = True

    def __init__(
        self,
        element_format: FloatingPoint,
        block_size=32,
    ):
        super().__init__()
        # check validity of format configuration
        assert isinstance(
            element_format, FloatingPoint
        ), "block format needs to be floating point"
        assert block_size > 0, f"block size has to be positive, got {block_size}"
        self.element_format = element_format
        self.scaler_format = FloatingPoint(
            mantissa=0,
            exponent=8,
            bias=127,
            unsigned=True,
        )
        self.block_size = block_size

        self.get_chunk_max = lambda chunk: torch.max(
            torch.abs(chunk), dim=-1, keepdim=True
        )[0]

    def cast(self, x: torch.Tensor, block_dim: int) -> torch.Tensor:
        _x = x.float().transpose(block_dim, -1)  # dim swap
        xshape = _x.shape  # remember shape
        _xs = torch.split(
            _x.reshape((-1, xshape[-1])), self.block_size, dim=-1
        )  # slice to chunks
        _ss = [
            2 ** torch.floor(torch.log2(self.get_chunk_max(chunk)))
            / self.element_format.largest_representable_power_of_two
            for chunk in _xs
        ]
        _x = torch.cat(
            [
                self.element_format.cast(chunk / scale) * scale
                for chunk, scale in zip(_xs, _ss)
            ],
            dim=block_dim,
        )  # quantize
        _x = _x.reshape(xshape).transpose_(block_dim, -1)  # recover shape
        return _x

    @property
    def bytes_per_elem(self) -> float:
        return (
            self.element_format.bytes_per_elem
            + self.scaler_format.bytes_per_elem / self.block_size
        )

    @property
    def bit_precision(self) -> float:
        return self.element_format.precision + 8.0 / self.block_size

    @classmethod
    def from_shorthand(cls, sh: str):
        conf = parse(
            "MXFP{precision:d}[E{exponent:d}M{mantissa:d}]{{{block_size:d}}}",
            sh,
        )
        assert conf["precision"] == conf["exponent"] + conf["mantissa"] + 1
        return cls(
            element_format=FloatingPoint(
                mantissa=conf["mantissa"],
                exponent=conf["exponent"],
                bias=2 ** (conf["exponent"] - 1) - 1,
                flush_subnormal=False,
                unsigned=False,
                rounding="nearest",
            ),
            block_size=conf["block_size"],
        )

    def __str__(self) -> str:
        return f"Simulated MXFP format: element format = {self.element_format}, scaler format = {self.scaler_format},\n block size = {self.block_size}"

    def __repr__(self) -> str:
        return f"MXFP{self.element_format.exponent+self.element_format.mantissa+1}[E{self.element_format.exponent}M{self.element_format.mantissa}]{{{self.block_size}}}"

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.element_format,
                self.block_size,
            ),
        )


class MXINT(BlockFloatingPoint):
    r"""
    This is a MXINT tensor format.
    """

    def __init__(
        self,
        precision=8,
        block_size=32,
    ):
        super().__init__(
            precision=precision,
            block_size=block_size,
            symmetric=True,
            rounding="nearest",
        )

    @classmethod
    def from_shorthand(cls, sh: str):
        conf = parse(
            "MXINT{precision:d}{{{block_size:d}}}",
            sh,
        )
        return cls(
            precision=conf["precision"],
            block_size=conf["block_size"],
        )

    def __str__(self) -> str:
        return f"Simulated MXINT format: precision bits = {self.precision}, block size = {self.block_size}\ncasting behavior: symmetric = {self.symmetric}, rounding = {self.rounding}"

    def __repr__(self) -> str:
        return f"MXINT{self.precision}{{{self.block_size}}}"

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.precision,
                self.block_size,
            ),
        )



import math
import torch
import skopt
from skopt import gp_minimize
import numpy as np
import copy
from typing import Optional
from functools import partial
from contextlib import contextmanager
from dmx.compressor.numerical.observer import HistogramObserver
from dmx.compressor.functional.approximate import NoApproximation


class LayerReconstructionMixin:
    r"""
    This mixin equips DmxModule with layer-reconstruction functionalities.
    Layer-reconstruction is any post-training process by which
    certain module parameters are fitted to optimize a local objective,
    usually by passing data (input activations) through the module.
    Examples are traditional static activation calibration,
    static SmoothQuant calibration, Optimal Brain Compression, etc.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.aft = None
        self.obc = None

    def _has_weight(self) -> bool:
        return "weight" in [n for n, _ in self.named_parameters()]

    def update_smoothquant_scale(self, input):
        if self.smoothquant is not None:
            self.smoothquant(input, self.effective_weight)

    def enable_quantizer_calib(self, state: bool, hyperparams) -> None:
        if hyperparams.inputs is not None:
            for _k in self.input_casts.keys():
                self.input_casts[_k].enable_calibration(
                    state, **hyperparams.inputs[_k].__dict__
                )
        if hyperparams.outputs is not None:
            for _k in self.output_casts.keys():
                self.output_casts[_k].enable_calibration(
                    state, **hyperparams.outputs[_k].__dict__
                )
        if self._has_weight():
            if hyperparams.weight is not None:
                self.weight_cast.enable_calibration(
                    state, **hyperparams.weight.__dict__
                )
            if hyperparams.weight_storage is not None:
                self.weight_storage_cast.enable_calibration(
                    state, **hyperparams.weight_storage.__dict__
                )

    def enable_smoothquant_calib(self, state: bool, hyperparams) -> None:
        if self.smoothquant is not None:
            if self.smoothquant.fused_to_weight[0] == 1:
                raise RuntimeError(
                    "SmoothQuant cannot be calibrated because it has been fused to weight already"
                )
            self.smoothquant.set_migration_strength(hyperparams.migration_strength)
            self.smoothquant.set_dynamic(False)  # only static needs calibration
            self.smoothquant.enable(not state)
            self.smoothquant.calibrating = state
            if not state and hyperparams.fuse_to_weight:
                self.smoothquant.fuse_to_weight(self.weight)

    def enable_optimal_brain_compression(self, state: bool, hyperparams) -> None:
        if isinstance(
            self,
            (
                torch.nn.Linear,
                torch.nn.Conv2d,
            ),
        ):
            if state:
                self.obc = OptimalBrainCompressor(self)
                # TODO: Weight Sparsifier should be handled.
                self.input_casts.disable_fake_quant()
                self.weight_cast.disable_fake_quant()
            else:
                self.input_casts.enable_fake_quant()
                self.weight_cast.enable_fake_quant()
                self.obc.apply(**hyperparams.__dict__)
                self.obc = None

    def enable_approximation_function_tuning(self, state: bool, hyperparams) -> None:
        if not isinstance(self.approximation_function, NoApproximation):
            if state:
                self.aft = ApproximationFunctionTuner(self, hyperparams.search_space)
            else:
                self.aft = None

    @contextmanager
    def calibrating_quantizers(self, hyperparams) -> None:
        self.enable_quantizer_calib(True, hyperparams)
        yield self
        self.enable_quantizer_calib(False, hyperparams)

    @contextmanager
    def calibrating_smoothquant(self, hyperparams) -> None:
        self.enable_smoothquant_calib(True, hyperparams)
        yield self
        self.enable_smoothquant_calib(False, hyperparams)

    @contextmanager
    def optimal_brain_compressing(self, hyperparams) -> None:
        self.enable_optimal_brain_compression(True, hyperparams)
        yield self
        self.enable_optimal_brain_compression(False, hyperparams)

    @contextmanager
    def tuning_approximation_function(self, hyperparams) -> None:
        self.enable_approximation_function_tuning(True, hyperparams)
        yield self
        self.enable_approximation_function_tuning(False, hyperparams)

    @contextmanager
    def slanc_tuning(self, hyperparams) -> None:
        if (
            isinstance(self, (torch.nn.LayerNorm, torch.nn.RMSNorm))
            and not isinstance(self.approximation_function, NoApproximation)
            and self.approximation_function.algorithm == "vsimd"
        ):
            if hyperparams.position == "post_attn":
                dev = hyperparams.device
                prev_ln_weight = hyperparams.prev_ln_weight.get_parameter("weight").detach().to(torch.float32)
                W_V = (
                    hyperparams.v_proj.get_parameter("weight")
                    .detach().to(torch.float32)
                )
                P = (
                    hyperparams.o_proj.get_parameter("weight")
                    .detach().to(torch.float32)
                )
                assert(P.shape[1] % W_V.shape[0] == 0)
                num_kv_heads = P.shape[1] // W_V.shape[0]
                norm = P @ (W_V.repeat(num_kv_heads, 1))
                assert norm.shape[0] == norm.shape[1]
                norm += torch.eye(norm.shape[0], device=dev)
                norm *= prev_ln_weight
                norm = torch.linalg.norm(norm, ord="fro")
            elif hyperparams.position == "post_mlp" \
                    and hyperparams.mlp_type == "standard":
                prev_ln_weight = hyperparams.prev_ln_weight.get_parameter("weight").detach().to(torch.float32)
                A = (
                    hyperparams.fc1.get_parameter("weight")
                    .detach().to(torch.float32)
                )
                B = (
                    hyperparams.fc2.get_parameter("weight")
                    .detach().to(torch.float32)
                )
                norm = (
                    torch.linalg.norm(prev_ln_weight, ord=1)
                    * torch.linalg.norm(A, ord=2)
                    * torch.linalg.norm(B, ord=2)
                    / prev_ln_weight.shape[0]
                )
            elif hyperparams.position == "post_mlp" \
                    and hyperparams.mlp_type == "llama":
                prev_ln_weight = hyperparams.prev_ln_weight.get_parameter("weight").detach().to(torch.float32)
                W_gate = (
                        hyperparams.gate_proj.get_parameter('weight')
                        .detach().to(torch.float32)
                        )
                W_up = (
                        hyperparams.up_proj.get_parameter('weight')
                        .detach().to(torch.float32)
                        )
                W_down = (
                        hyperparams.down_proj.get_parameter('weight')
                        .detach().to(torch.float32)
                        )
                norm = torch.linalg.norm(W_down @ (W_up * prev_ln_weight), 'fro') \
                        * torch.linalg.norm(W_gate * prev_ln_weight, ord=2)
            elif hyperparams.position == "first":
                norm = 1.0
            #Since the approximator is shared between all instances of same module
            #deepcopy to allow each simd module to have different simd parameters
            self.approximator.function = copy.deepcopy(self.approximator.function)

            #SLANC assumes the layernorm input will be divided by norm
            #However, the SIMD kernels multiply the input by the norm parameter
            #Hence the 1/x transformation
            self.approximator.function.extra_params.update({"norm": 1.0/norm})
        yield self


class ApproximationFunctionTuner:
    def __init__(self, module, search_space):
        self.module = module
        self.search_space = search_space
        self.solver = partial(gp_minimize, n_calls=20)

    def optimize(self, input, *args, **kwargs):
        self.module.approximator.function = copy.deepcopy(
            self.module.approximator.function
        )
        module_aft = self.module.aft
        # To avoid infinite recursion in the module's forward pass
        self.module.aft = None

        @skopt.utils.use_named_args(self.search_space)
        def obj_func(**extra_params):
            self.module.approximator.function.extra_params.update(extra_params)
            _ = self.module(input, *args, **kwargs)
            _e = self.module.approximation_error
            return torch.nn.functional.mse_loss(_e, torch.zeros_like(_e)).item()

        _res = self.solver(obj_func, self.search_space)
        self.module.aft = module_aft
        self.module.approximator.function.extra_params.update(
            {_p.name: _opt for _p, _opt in zip(self.search_space, _res.x)}
        )


class OptimalBrainCompressor:
    H: Optional[torch.Tensor] = None

    def __init__(self, module):
        self.module = module
        self.example_counter = 0

    def measure_hessian(self, inp):
        # TODO: clean up ugliness
        if self.H is None:
            self.H = torch.zeros(1).to(inp.device)
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if isinstance(self.module, torch.nn.Linear):
            if len(inp.shape) == 3:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()
        if isinstance(self.module, torch.nn.Conv2d):
            inp = torch.nn.functional.unfold(
                inp,
                self.module.kernel_size,
                dilation=self.module.dilation,
                padding=self.module.padding,
                stride=self.module.stride,
            )
            inp = inp.permute([1, 0, 2])
            inp = inp.flatten(1)
        self.H *= self.example_counter / (self.example_counter + tmp)
        self.example_counter += tmp
        inp = math.sqrt(2 / self.example_counter) * inp.float()
        self.H = self.H + inp.matmul(inp.t())

    def apply(self, microblock_size=1, block_size=128, percdamp=0.01):
        assert block_size % microblock_size == 0
        if self.module.weight_sparsifier.sparseness.blocked:
            assert (
                microblock_size % self.module.weight_sparsifier.sparseness.block_size
                == 0
            )
        if self.module.weight_cast.format.blocked:
            assert microblock_size % self.module.weight_cast.format.block_size == 0

        # TODO: clean up ugliness
        W = self.module.weight.data.clone()
        if isinstance(self.module, torch.nn.Conv2d):
            W = W.flatten(1)
        W = W.float()
        ncols = W.shape[1]

        H = self.H
        del self.H
        dead = torch.diag(H) == 0
        H[dead, dead] = 1
        W[:, dead] = 0

        Q = torch.zeros_like(W)

        damp = percdamp * torch.mean(torch.diag(H))
        diag = torch.arange(ncols, device=H.device)
        H[diag, diag] += damp
        H = torch.linalg.cholesky(H)
        H = torch.cholesky_inverse(H)
        H = torch.linalg.cholesky(H, upper=True)
        Hinv = H

        # TODO: allow OBC candidate selection

        for i1 in range(0, ncols, block_size):
            i2 = min(i1 + block_size, ncols)
            count = i2 - i1

            _W = W[:, i1:i2].clone()
            _Q = torch.zeros_like(_W)
            _E = torch.zeros_like(_W)
            _Hinv = Hinv[i1:i2, i1:i2]

            for j1 in range(0, count, microblock_size):
                j2 = min(j1 + microblock_size, count)
                w = _W[:, j1:j2]
                hinv = _Hinv[j1:j2, j1:j2]
                q = self.module.weight_hypernet(w)
                err = (w - q).matmul(torch.linalg.inv(hinv))
                _Q[:, j1:j2] = q
                _W[:, j2:] -= err.matmul(_Hinv[j1:j2, j2:])
                _E[:, j1:j2] = err

            Q[:, i1:i2] = _Q
            W[:, i2:] -= _E.matmul(Hinv[i1:i2, i2:])

        torch.cuda.synchronize()

        self.module.weight.data = Q.reshape(self.module.weight.shape).to(
            self.module.weight.data.dtype
        )



import torch
from dmx.compressor import nn
import pytest

torch.manual_seed(0)


def test_linear():
    linear = nn.Linear(64, 64)
    gm = torch.fx.GraphModule(linear, linear.to_compiler_graph())
    inp = torch.rand((1, 64))
    out0 = linear(inp)
    out1 = gm(inp)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_resadd():
    resadd = nn.ResAdd()
    gm = torch.fx.GraphModule(resadd, resadd.to_compiler_graph())
    inp = torch.rand((1, 64))
    out0 = resadd(inp, inp)
    out1 = gm(inp, inp)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_mul():
    mul = nn.Mul()
    gm = torch.fx.GraphModule(mul, mul.to_compiler_graph())
    inp = torch.rand((1, 64))
    out0 = mul(inp, inp)
    out1 = gm(inp, inp)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_actactmatmul():
    actactmatmul = nn.ActActMatMul()
    gm = torch.fx.GraphModule(actactmatmul, actactmatmul.to_compiler_graph())
    inp1 = torch.rand((1, 64))
    inp2 = torch.rand((64, 4))
    out0 = actactmatmul(inp1, inp2)
    out1 = gm(inp1, inp2)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_exp():
    _exp = nn.Exp()
    gm = torch.fx.GraphModule(_exp, _exp.to_compiler_graph())
    inp1 = torch.rand((1, 64))
    out0 = _exp(inp1)
    out1 = gm(inp1)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_baddbmm():
    baddbmm = nn.BAddBMM()
    gm = torch.fx.GraphModule(baddbmm, baddbmm.to_compiler_graph())
    inp1 = torch.rand((10, 3, 4))
    inp2 = torch.rand((10, 3, 64))
    inp3 = torch.rand((10, 64, 4))
    out0 = baddbmm(inp1, inp2, inp3)
    out1 = gm(inp1, inp2, inp3)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_embed():
    embed = nn.Embedding(10, 64)
    gm = torch.fx.GraphModule(embed, embed.to_compiler_graph())
    inp = torch.randint(0, 10, (1, 10))
    out0 = embed(inp)
    out1 = gm(inp)
    assert torch.allclose(out0, out1, atol=1e-5)


@pytest.mark.parametrize(
    "module_type",
    (
        "RMSNorm",
        "LayerNorm",
        "GemmaRMSNorm",
    ),
)
def test_norms(module_type):
    norm = eval(f"nn.{module_type}(64)")
    gm = torch.fx.GraphModule(norm, norm.to_compiler_graph())
    inp = torch.rand((1, 64))
    out0 = norm(inp)
    out1 = gm(inp)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_groupnorm():
    groupnorm = nn.GroupNorm(8, 64)
    gm = torch.fx.GraphModule(groupnorm, groupnorm.to_compiler_graph())
    inp = torch.rand((1, 64))
    out0 = groupnorm(inp)
    out1 = gm(inp)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_dropout():
    dropout = nn.Dropout(0)
    gm = torch.fx.GraphModule(dropout, dropout.to_compiler_graph())
    inp = torch.rand((1, 64))
    out0 = dropout(inp)
    out1 = gm(inp)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_clippedgelu():
    act = nn.ClippedGELU(0, 1)
    gm = torch.fx.GraphModule(act, act.to_compiler_graph())
    inp = torch.rand((1, 64))
    out0 = act(inp)
    out1 = gm(inp)
    assert torch.allclose(out0, out1, atol=1e-5)


@pytest.mark.parametrize(
    "module_type",
    (
        "Softmax",
        "ReLU",
        "ReLU6",
        "SiLU",
        "Tanh",
        "NewGELU",
        "FastGELU",
        "QuickGELU",
        "BloomGELU",
        "GELU",
    ),
)
def test_activations(module_type):
    act = eval(f"nn.{module_type}()")
    gm = torch.fx.GraphModule(act, act.to_compiler_graph())
    inp = torch.rand((1, 64))
    out0 = act(inp)
    out1 = gm(inp)
    assert torch.allclose(out0, out1, atol=1e-5)


def test_apply_rope():
    rope = nn.ApplyRotaryPosEmb()
    q = torch.randn(1, 64, 64)
    k = torch.randn(1, 64, 64)
    cos = torch.randn(1, 64)
    sin = torch.randn(1, 64)

    out0 = rope(q, k, cos, sin)
    qdq = rope.to_compiler_graph()
    qdq_module = torch.fx.GraphModule(rope, qdq)
    out1 = qdq_module(q, k, cos, sin)
    assert torch.allclose(out0[0], out1[0], atol=1e-5)
    assert torch.allclose(out0[1], out1[1], atol=1e-5)



import torch
from types import SimpleNamespace


__ALL__ = ["experimental"]

class experimental(SimpleNamespace):
    def silu(
        input: torch.Tensor, inplace: bool = False, **extra_params
    ) -> torch.Tensor:
        r"""
        An absurd approximation of the silu function
        """
        # 1. parse extra_params
        _scale = extra_params["scale"]
        # 2. resolve paramter conflicts
        assert (
            not inplace
        ), "inplace has to be False, not functionally meaningful anyway"
        # 3. custom logic
        return torch.nn.functional.relu(input.to(torch.float16)) * _scale



import pytest

import torch
from dmx.compressor.sparse import Sparsify, Sparseness

RANDOM_SEED = 0

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
torch.manual_seed(RANDOM_SEED)


@pytest.mark.parametrize(
    "tensor_shape",
    (
        (1024, 1024),
        (256, 256, 32, 32),
        (8, 256, 256),
    ),
)
@pytest.mark.parametrize(
    "sparseness",
    (
        "TOPK{0.5}(U)",
        "TOPK{0.5}(M)",
        "BTOPK{4:8,-1}(U)",
        "BTOPK{4:8,-1}(M)",
        "BTOPK{2:8,-1}(U)",
        "BTOPK{2:8,-1}(M)",
        "BERN",
    ),
)
@pytest.mark.parametrize(
    "backward_mode",
    (
        "STE",
        "supermask",
        "joint",
    ),
)
def test_sparsify(tensor_shape, sparseness, backward_mode):
    """Test that `Sparsify` produces a correct sparseness pattern,
    as evidenced by gradients of weights and scores."""
    sp = Sparsify(tensor_shape, sparseness, backward_mode).to(device)
    x = torch.randn(tensor_shape, requires_grad=True, device=device)
    y = sp(x)
    y.backward(torch.ones_like(y))

    if backward_mode == "STE":
        assert isinstance(x.grad, torch.Tensor) and sp.score.grad is None
    elif backward_mode == "supermask":
        assert x.grad is None and isinstance(sp.score.grad, torch.Tensor)
    elif backward_mode == "joint":
        assert isinstance(x.grad, torch.Tensor) and isinstance(
            sp.score.grad, torch.Tensor
        )


@pytest.mark.parametrize(
    "tensor_shape",
    (
        (1024, 1024),
        (256, 256, 32, 32),
        (8, 256, 256),
    ),
)
def test_transformation(tensor_shape):
    """Test updating of sparseness and backward mode."""
    sp = Sparsify(tensor_shape).to(device)
    assert repr(sp.sparseness) == "DENSE"
    assert sp.backward_mode == "STE"

    sp.configure(
        sparseness="BERN",
        backward_mode="supermask",
        score_func=lambda score, input: score,
    )
    assert repr(sp.sparseness) == "BERN"
    assert sp.backward_mode == "supermask"

    sp.sparseness = Sparseness.from_shorthand("DENSE")
    sp.backward_mode = "STE"
    assert repr(sp.sparseness) == "DENSE"
    assert sp.backward_mode == "STE"

    sp.configure(
        sparseness="TOPK{0.5}(U)",
        backward_mode="joint",
        score_func=lambda score, input: torch.abs(input),
    )
    assert repr(sp.sparseness) == "TOPK{0.5}(U)"
    assert sp.backward_mode == "joint"

    sp.sparseness = Sparseness.from_shorthand("BTOPK{4:8,-1}(U)")
    sp.backward_mode = "STE"
    assert repr(sp.sparseness) == "BTOPK{4:8,-1}(U)"
    assert sp.backward_mode == "STE"



import pytest
import torch
import torch.nn as nn
import torch.nn.functional as F
from dmx.compressor import format
from dmx.compressor import nn as dmxnn
from dmx.compressor.modeling import DmxModel, DmxConfigRule, DmxModuleConfig
from dmx.compressor.numerical.format import Same
from dmx.compressor.sparse import Dense, Sparseness
from dmx.compressor.advanced_recipe import (
    DmxModuleSmoothQuantHyperparams,
    DmxSmoothQuantRecipe,
)

RANDOM_SEED = 0

torch.manual_seed(RANDOM_SEED)
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class Lenet5(nn.Module):
    r"""
    Taken from https://pytorch.org/tutorials/beginner/blitz/neural_networks_tutorial.html
    """

    def __init__(self):
        super().__init__()
        # 1 input image channel, 6 output channels, 5x5 square convolution
        # kernel
        self.conv1 = nn.Conv2d(1, 6, 5)
        self.conv2 = nn.Conv2d(6, 16, 5)
        # an affine operation: y = Wx + b
        self.fc1 = nn.Linear(16 * 5 * 5, 120)  # 5*5 from image dimension
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)
        # Max pooling over a (2, 2) window
        self.mp1 = nn.MaxPool2d((2, 2))
        # If the size is a square, you can specify with a single number
        self.mp2 = nn.MaxPool2d(2)

    def forward(self, x):
        x = self.mp1(F.relu(self.conv1(x)))
        x = self.mp2(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)  # flatten all dimensions except the batch dimension
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


conv_quantize_input = DmxConfigRule(
    module_types=(dmxnn.Conv2d,),
    module_config=DmxModuleConfig(input_format=format.BFP12_128),
)
conv_quantize_weight = DmxConfigRule(
    module_types=(dmxnn.Conv2d,),
    module_config=DmxModuleConfig(weight_format=format.BFP12_128),
)
conv_sparsify_weight = DmxConfigRule(
    module_types=(dmxnn.Conv2d,),
    module_config=DmxModuleConfig(
        weight_sparseness=Sparseness.from_shorthand("TOPK{0.5}(U)")
    ),
)
fc_quantize_input = DmxConfigRule(
    module_types=(dmxnn.Linear,),
    module_config=DmxModuleConfig(input_format=format.BFP12_128),
)
fc_quantize_weight = DmxConfigRule(
    module_types=(dmxnn.Linear,),
    module_config=DmxModuleConfig(weight_format=format.BFP12_128),
)
fc_sparsify_weight = DmxConfigRule(
    module_types=(dmxnn.Linear,),
    module_config=DmxModuleConfig(
        weight_sparseness=Sparseness.from_shorthand("TOPK{0.5}(U)")
    ),
)


def _create_input(batch_size: int = 1):
    return torch.randn(batch_size, 1, 32, 32)


def _create_model(transformations):
    _model = DmxModel.from_torch(Lenet5())
    conv_quantize_input.apply_to(_model)
    fc_quantize_input.apply_to(_model)
    for tr in transformations:
        tr.apply_to(_model)
    hp_gen = lambda __model: {
        _m: DmxModuleSmoothQuantHyperparams()
        for _, _m in _model.named_dmx_modules()
        if isinstance(_m, (nn.Linear, nn.Conv2d))
    }
    with DmxSmoothQuantRecipe(hp_gen).applied_to(_model), torch.no_grad():
        _model(_create_input(batch_size=128))
    _model.eval()
    return _model


def _check_if_folded(model: DmxModel) -> bool:
    return all(
        [
            all(
                [
                    m.bias_cast is None or isinstance(m.bias_cast.format, Same),
                    m.weight_sparsifier is None
                    or isinstance(m.weight_sparsifier.sparseness, Dense),
                    m.smoothquant is None or m.smoothquant.fused_to_weight[0] == 1,
                    m.weight_cast is None or isinstance(m.weight_cast.format, Same),
                ]
            )
            for _, m in model.named_dmx_modules()
        ]
    )


@pytest.mark.parametrize(
    "dmx_transformations",
    (
        (),
        (conv_quantize_weight,),
        (fc_quantize_weight,),
        (conv_sparsify_weight,),
        (fc_sparsify_weight,),
        (conv_quantize_weight, fc_quantize_weight),
        (conv_sparsify_weight, fc_sparsify_weight),
        (conv_quantize_weight, conv_sparsify_weight),
        (fc_quantize_weight, fc_sparsify_weight),
        (
            conv_quantize_weight,
            fc_quantize_weight,
            conv_sparsify_weight,
            fc_sparsify_weight,
        ),
    ),
)
def test_fold_weights_and_biases(dmx_transformations):
    x = _create_input()
    model = _create_model(dmx_transformations)
    with torch.no_grad():
        y_ref = model(x)
    model.fold_weights_and_biases()
    assert _check_if_folded(model)
    with torch.no_grad():
        y = model(x)
    assert torch.all(y == y_ref)
